/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Luigi Zangrando <zangrando@pd.infn.it>
 *
 */

package org.glite.ce.monitor;

import java.net.URI;
import java.rmi.RemoteException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import javax.naming.Context;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;

import org.apache.axis.MessageContext;
import org.apache.log4j.Logger;
import org.glite.ce.commonj.utils.CEUtils;
import org.glite.ce.commonj.utils.Timer;
import org.glite.ce.commonj.utils.TimerTask;
import org.glite.ce.faults.AuthenticationFault;
import org.glite.ce.faults.AuthorizationFault;
import org.glite.ce.faults.BaseFaultType;
import org.glite.ce.faults.GenericFault;
import org.glite.ce.monitor.holder.ActionHolder;
import org.glite.ce.monitor.holder.NotificationHolder;
import org.glite.ce.monitor.holder.PredefinedSubscriptionHolder;
import org.glite.ce.monitor.holder.QueryProcessorHolder;
import org.glite.ce.monitor.holder.SensorEventArrayList;
import org.glite.ce.monitor.holder.SensorHolder;
import org.glite.ce.monitor.holder.TopicHolder;
import org.glite.ce.monitor.holder.TopicNotFoundException;
import org.glite.ce.monitor.registry.SubscriptionRegistry;
import org.glite.ce.monitorapij.faults.DialectNotSupportedFault;
import org.glite.ce.monitorapij.faults.SubscriptionFault;
import org.glite.ce.monitorapij.faults.SubscriptionNotFoundFault;
import org.glite.ce.monitorapij.faults.TopicNotSupportedFault;
import org.glite.ce.monitorapij.queryprocessor.QueryProcessor;
import org.glite.ce.monitorapij.resource.Resource;
import org.glite.ce.monitorapij.resource.ResourceEvent;
import org.glite.ce.monitorapij.resource.ResourceHolder;
import org.glite.ce.monitorapij.resource.ResourceListener;
import org.glite.ce.monitorapij.resource.types.Action;
import org.glite.ce.monitorapij.resource.types.Dialect;
import org.glite.ce.monitorapij.resource.types.Property;
import org.glite.ce.monitorapij.resource.types.SubscriptionPersistent;
import org.glite.ce.monitorapij.sensor.Sensor;
import org.glite.ce.monitorapij.sensor.SensorEvent;
import org.glite.ce.monitorapij.sensor.SensorException;
import org.glite.ce.monitorapij.sensor.SensorListener;
import org.glite.ce.monitorapij.sensor.SensorOutputDataFormat;
import org.glite.ce.monitorapij.types.Event;
import org.glite.ce.monitorapij.types.Info;
import org.glite.ce.monitorapij.types.Notification;
import org.glite.ce.monitorapij.types.Subscription;
import org.glite.ce.monitorapij.types.SubscriptionRef;
import org.glite.ce.monitorapij.types.Topic;
import org.glite.ce.monitorapij.ws.CEMonitorPort;
import org.glite.security.util.FileCertReader;

/**
 * The <code>CEMonitorService</code> represents the core of the monitor
 * service.<br>
 * It implements three interfaces: <code>CEMonitorPort</code> for functions
 * offered via Web Service, <code>SensorListener</code> to manage events
 * coming from sensor plugins, and <code>ResourceListener</code> to manage
 * events coming from all monitor resources.
 * <p>
 * When creating an instance of this class it provides initialization for all
 * the holders of CE framework: <code>ActionHolder</code>,
 * <code>NotificationHolder</code>, <code>PredefinedSubscriptionHolder</code>,
 * <code>QueryProcessorHolder</code>, <code>SensorHolder</code>,
 * <code>TopicHolder</code> and <code>SubscriptionRegistry</code>.<br>
 * It triggers all these holders to manage, from the top level, the totality of
 * the occurring events and relative notifications, both synchronous or
 * asynchronous.<br>
 * Moreover it controls the mechanism to check if new plugins are added or known
 * ones are modified.
 * <p>
 * All these tasks are controlled by sceduling subcomponents execution and
 * managing the relative communications.<br>
 * The initialization of the service is operated by
 * <code>MonitorContextListener</code> class.
 */
public class CEMonitorService implements CEMonitorPort, SensorListener, ResourceListener {
    private final static Logger logger = Logger.getLogger(CEMonitorService.class.getName());
    private static boolean notificationSharedThread = true;
    public static MessageContext messageContext;
    protected static SubscriptionRegistry subscriptionRegistry = null;
    protected static TopicHolder topicHolder = null;
    protected static SensorHolder sensorHolder = null;
    protected static ActionHolder actionHolder = null;
    protected static QueryProcessorHolder queryProcessorHolder = null;
    protected static PredefinedSubscriptionHolder predefinedSubscriptionHolder = null;
    protected static Hashtable<String, NotificationHolder> notificationHolderList = null;
    protected static Hashtable<String, String> visibilityTable;
    protected static Timer timer = null;
    protected static boolean initialized = false;
    protected static Info info;
    private static int maxExpirationTime_Hours = 8760; //hours for year

    /**
     * Creates a new CEMonitorService object. If the axis
     * <code>MessageContext</code> hasn't been initialized yet the current
     * context is called.
     */
    public CEMonitorService() {
    	if(messageContext == null) {
    		messageContext = MessageContext.getCurrentContext();
    	}
    	
    	if(visibilityTable == null) {
    		visibilityTable = new Hashtable(0);
    		visibilityTable.put("HIGH", "ALL");
    		visibilityTable.put("MEDIUM", "GROUP");
    		visibilityTable.put("LOW", "USER");
            
          //  MemoryMXBean mbean = ManagementFactory.getMemoryMXBean();
           // mbean.setVerbose(true);
    	}
    }

    /**
     * Initialize the monitor service. This method is called when the
     * <code>CEMonitorService</code> class is instantiated. It triggers the
     * initialization of all the holders of the service: TopicHolder,
     * ActionHolder, SensorHolder, SubscriptionRegistry, QueryProcessorHolder,
     * PredefinedSubscriptionHolder.
     */
    public void init() throws Exception {
        if (initialized) {
            return;
        }

        logger.info("Initializing CEMon...");
        
        org.glite.ce.monitorapij.types.Property[] property = new org.glite.ce.monitorapij.types.Property[1];
        property[0] = new org.glite.ce.monitorapij.types.Property("hostDN", "N/A");
        logger.debug("init() - Got hostDN property=[" + property[0].getValue() + "]");

        info = new Info();
        info.setDescription("CEMon service");
        info.setInterfaceVersion("N/A");
        info.setServiceVersion("N/A");
        info.setStartupTime(Calendar.getInstance());
        info.setProperty(property);

        try {
            String confProvider = CEUtils.getConfigurationProvider(CEUtils.CEMONITOR_LABEL);
            String confProviderURL = CEUtils.getConfigurationURL(CEUtils.CEMONITOR_LABEL);

            if (confProvider == null) {
                throw new Exception("Configuration provider not defined!");
            }

            if (confProviderURL == null) {
                throw new Exception("Configuration provider URL not defined!");
            }

            Hashtable<String, String> jndiProperties = new Hashtable<String, String>(2);
            jndiProperties.put(Context.INITIAL_CONTEXT_FACTORY, confProvider);
            jndiProperties.put(Context.PROVIDER_URL, confProviderURL);

            DirContext rootContext = (DirContext) new InitialDirContext(jndiProperties);

            Attributes attribute = rootContext.getAttributes("", new String[] { "maxSubscriptionExpirationTime_hours", 
                    "sslcertfile", "ServiceDescription", "ServiceVersion", "InterfaceVersion", "notificationSharedThread" });

            if (attribute != null) {
                if (attribute.get("maxSubscriptionExpirationTime_hours") != null) {
                    String expirationTime = (String) attribute.get("maxSubscriptionExpirationTime_hours").get(0);
                    if (expirationTime != null) {
                        try {
                            maxExpirationTime_Hours = Integer.parseInt(expirationTime);
                        } catch (Exception e) {
                            logger.warn("init() - Parameter \"maxSubscriptionExpirationTime_hours\" not found: using the default value (" + maxExpirationTime_Hours
                                    + " hours). Follow the Stack trace:", e);
                        }
                    } else {
                        logger.warn("init() - \"maxSubscriptionExpirationTime_hours\" parameter not found!. Using the default value (" + maxExpirationTime_Hours + " hours)");
                    }
                }

                if (attribute.get("sslcertfile") != null) {
                    String filename = (String) attribute.get("sslcertfile").get(0);

                    try {
                        FileCertReader reader = new FileCertReader();
                        Vector allCerts = reader.readCerts(filename);
                        X509Certificate cert = (X509Certificate) allCerts.elementAt(0);
                        String hostDN = "/" + cert.getSubjectDN().toString().replace(',', '/');
                        int index = hostDN.indexOf("/E=");
                        if (index > 0) {
                            hostDN = hostDN.substring(0, index) + "/Email=" + hostDN.substring(index + 3);
                            logger.debug("init() - Mangled hostDN. New value is [" + hostDN + "]");
                        }
                        property[0].setValue(hostDN);
                    } catch (CertificateException e) {
                        logger.warn("init() - Parameter \"sslcertfile\" not found!. Follow the Stack trace:", e);
                    }
                }

                if (attribute.get("ServiceDescription") != null) {
                    info.setDescription((String) attribute.get("ServiceDescription").get(0));
                }

                if (attribute.get("ServiceVersion") != null) {
                    info.setServiceVersion((String) attribute.get("ServiceVersion").get(0));
                }

                if (attribute.get("InterfaceVersion") != null) {
                    info.setInterfaceVersion((String) attribute.get("InterfaceVersion").get(0));
                }

                if (attribute.get("notificationSharedThread") != null) {
                    notificationSharedThread = Boolean.parseBoolean((String) attribute.get("notificationSharedThread").get(0));
                }
            }
        } catch (NamingException nEx) {
            logger.error("init() - Catched NamingException. Follow the Stack trace:", nEx);
        }

        logger.debug("init() - Initialized info object...");
        logger.debug("init() - Description=[CEMon service]");
        logger.debug("init() - InterfaceVersion=[N/A]");
        logger.debug("init() - ServiceVersion=[N/A]");
        logger.debug("init() - StartupTime=[" + Calendar.getInstance().getTime().toString() + "]");
        logger.debug("init() - property=[" + property[0].getValue() + "]");

        if (timer == null) {
            timer = new Timer("CEMonitorService_TIMER", true);
        }
        
        initNotificationHolder();
        initQueryProcessorHolder();
        initActionHolder();
        initSensorHolder();
        initTopicHolder();
        initSubscriptionRegistry();
        initPredefinedSubscriptionHolder();

        initialized = true;

        logger.info("Initializing CEMon... done!");
    }


    public void destroy() {
        logger.info("destroy() - destroying service... ");
        
        for (NotificationHolder notificationHolder : notificationHolderList.values()) {        
            notificationHolder.destroy();
        }
        
        if (timer != null) {
            logger.info("destroy() - cancelling timer");
            timer.cancel();
        }

        if (sensorHolder != null) {
            logger.info("destroy() - destroying sensors");
            Sensor[] sensors = sensorHolder.getSensors();
            for (int i = 0; i < sensors.length; i++) {
                sensors[i].destroySensor();
            }
        }

        if (queryProcessorHolder != null) {
            logger.info("destroy() - destroying queryProcessorHolder");
            queryProcessorHolder.cancel();
        }

        if (actionHolder != null) {
            logger.info("destroy() - destroying actionHolder");
            actionHolder.cancel();
        }

        if (predefinedSubscriptionHolder != null) {
            logger.info("destroy() - destroying predefinedSubscriptionHolder");
            predefinedSubscriptionHolder.cancel();
        }

        if (topicHolder != null) {
            logger.info("destroy() - destroying topic holder");
            topicHolder.destroy();
        }
        
        logger.info("destroy() - destroying service... done!");
    }


    /**
     * Initialize the NotificationHolder creating an empty list of notifications
     * and instantiating the timer.
     */
    protected void initNotificationHolder() {
        logger.debug("initNotificationHolder() - executing method");
        
        if (notificationHolderList == null) {
            logger.info("Initializing NotificationHolder...");
            
            notificationHolderList = new Hashtable<String, NotificationHolder>(0);
           
            logger.info("Initializing NotificationHolder... done!");
        }
        
        logger.debug("initNotificationHolder() - executed");
    }

    /**
     * Initialize the subscription registry. <br>
     * In this registry must be opportunely saved informations about persistent
     * subscriptions. These must be initially activated without an explicit
     * request and eventually restored.
     * <p>
     * The message context must contain the correct path to a file specifying
     * these registry informations. If this path is not correctly specified a
     * default path is checked to find the registry file. This path is
     * "home.dir"/subscriptions/subscription.reg.
     * 
     * @throws Exception
     *             This exception is thrown if the registry file does not
     *             contain the subscriptions or is not formatted as expected.
     */
    protected void initSubscriptionRegistry() throws Exception {
        logger.debug("initSubscriptionRegistry() - executing method");
        
        if (subscriptionRegistry == null) {
            logger.info("Initializing SubscriptionRegistry...");
            
            String confProvider = CEUtils.getConfigurationProvider(CEUtils.CEMONITOR_LABEL);
            String confProviderURL = CEUtils.getConfigurationURL(CEUtils.CEMONITOR_LABEL);
            String cachePath = "/opt/glite/var/cache";
            int cacheSize = 100;

            if (confProvider == null) {
                logger.error("initSubscriptionRegistry() - confProvider not defined");

                throw new Exception("initSubscriptionRegistry() - confProvider not defined");
            }

            if (confProviderURL == null) {
                logger.error("initSubscriptionRegistry() - confProviderURL not defined");

                throw new Exception("initSubscriptionRegistry() - confProviderURL not defined");
            }

            // if( confProvider!=null && confProviderURL!=null ) {
            Hashtable<String, String> jndiProperties = new Hashtable<String, String>(2);
            jndiProperties.put(Context.INITIAL_CONTEXT_FACTORY, confProvider);
            jndiProperties.put(Context.PROVIDER_URL, confProviderURL);

            String subscriptionRegistryConfProvider = "org.glite.ce.commonj.jndi.provider.fscachedprovider.CEGeneralDirContextFactory";

            try {
                DirContext rootContext = (DirContext) new InitialDirContext(jndiProperties);
                Attributes serviceAttrs = rootContext.getAttributes("");
                Attribute attr = serviceAttrs.get("backendLocation");
                if (attr != null) {
                    cachePath = (String) attr.get(0);
                }

                attr = serviceAttrs.get("backendCacheSize");
                if (attr != null) {
                    cacheSize = Integer.parseInt((String) attr.get(0));
                }

                if (serviceAttrs.get("subscriptionRegistryConfProvider") == null) {
                    logger.error("initSubscriptionRegistry() - subscriptionRegistryConfProvider NOT defined in the configuration file. Using the default ["
                            + subscriptionRegistryConfProvider + "]");
                } else {
                    subscriptionRegistryConfProvider = (String) serviceAttrs.get("subscriptionRegistryConfProvider").get(0);
                }

            } catch (NamingException nEx) {
                logger.error("init() - Catched NamingException. Follow the Stack trace:", nEx);
            }

            logger.debug("initSubscriptionRegistry() - Initialized configuration: confProvider=[" + confProvider + "] - " + "confProviderURL=[" + confProviderURL + "] - "
                    + "cachePath=[" + cachePath + "] - " + "cacheSize=[" + cacheSize + "]");

            subscriptionRegistry = new SubscriptionRegistry(cachePath + "/subscription", cacheSize, subscriptionRegistryConfProvider /* confProvider */);

            SubscriptionPersistent[] subscription = subscriptionRegistry.getAllSubscriptions();

            logger.debug("initSubscriptionRegistry() - " + "Found [" + subscription.length + "] subscription(s)");

            for (int i = 0; i < subscription.length; i++) {
                if (!subscription[i].isExpired()) {
                    addToNotificationHolder(subscription[i]);
                } else {
                    logger.debug("initSubscriptionRegistry() - UNREGISTERING EXPIRED SUBSCRIPTION: " + subscription[i].getCompleteInfo());

                    subscriptionRegistry.unregister(subscription[i]);
                }
            }

            logger.info("Initializing SubscriptionRegistry... done!");
        }
        logger.debug("initSubscriptionRegistry() - executed.");
    }



    /**
     * Initialize the <code>QueryProcessorHolder</code> searching for query
     * processors and relative configuration file.<br>
     * If properties "queryProcessorURI" and "queryProcessorConfigURI" are not
     * correctly defined in the context the default paths will be
     * "home.dir"/query_processors and "home.dir"/queryProcessorConfig.xml.<br>
     * Then the monitor is subscribed as a listener for queryProcessor-related
     * events.<br>
     * Finally the execution of the <code>QueryProcessorHolder</code> is
     * scheduled by setting the timer. The holder will be executed the first
     * time with a delay of one second, then repeatedly executed each minute.
     * 
     * @throws Exception
     *             This exception is thrown if the query processor config file
     *             does not contain the correct informations or is not formatted
     *             as expected.
     */
    protected void initQueryProcessorHolder() throws Exception {
        logger.debug("initSubscriptionRegistry() - executing method");
        
        if (queryProcessorHolder == null) {
            logger.info("Initializing QueryProcessorHolder...");
            
            queryProcessorHolder = new QueryProcessorHolder();
            queryProcessorHolder.addResourceListener(this);
            timer.schedule(queryProcessorHolder, 0, 600000, TimerTask.EXECUTION_TYPE.FIXED_DELAY_POST_EXECUTION);
            
            logger.info("Initializing QueryProcessorHolder... done!");
        }

        logger.debug("initQueryProcessorHolder() - executed.");
    }



    /**
     * Initialize the <code>ActionHolder</code> searching for actions plugins
     * and relative configuration file.<br>
     * If properties "actionURI" and "actionConfigURI" are not correctly defined
     * in the context, the default paths will be "home.dir"/actions and
     * "home.dir"/config/actionConfig.xml.<br>
     * Then the monitor is subscribed as a listener for action-related events.<br>
     * Finally the execution of the <code>ActionHolder</code> is scheduled by
     * setting the timer. The holder will be executed the first time with a
     * delay of two seconds, then repeatedly executed each minute.
     * 
     * @throws Exception
     *             This exception is thrown if the action config file does not
     *             contain the correct informations or is not formatted as
     *             expected.
     */
    protected void initActionHolder() throws Exception {
        logger.debug("initActionHolder() - executing method");
        
        if (actionHolder == null) {
            logger.info("Initializing ActionHolder...");
            
            actionHolder = new ActionHolder();
            actionHolder.addResourceListener(this);
            timer.schedule(actionHolder, 1000, 600000, TimerTask.EXECUTION_TYPE.FIXED_DELAY_POST_EXECUTION);

            logger.info("Initializing ActionHolder... done!");
        }

        logger.debug("initActionHolder() - executed.");
    }

    

    /**
     * Initialize the <code>SensorHolder</code> searching for sensor plugins
     * and relative configuration file. If properties "sensorURI" and
     * "sensorConfigURI" are not correctly defined in the context the default
     * paths will be "home.dir"/sensors and "home.dir"/config/sensorConfig.xml.<br>
     * Then the monitor is subscribed as a listener for sensor-related events.<br>
     * Finally the execution of the <code>SensorHolder</code> is scheduled by
     * setting the timer. The holder will be executed the first time with a
     * delay of three seconds, then repeatedly executed each minute.
     * 
     * @throws Exception
     *             This exception is thrown if the sensor config file does not
     *             contain the correct informations or is not formatted as
     *             expected.
     */
    protected void initSensorHolder() throws Exception {
        logger.debug("initSensorHolder() - executing method");
        
        if (sensorHolder == null) {
            logger.info("Initializing SensorHolder...");
            
            sensorHolder = new SensorHolder();
            sensorHolder.addResourceListener(this);
            timer.schedule(sensorHolder, 2000, 600000, TimerTask.EXECUTION_TYPE.FIXED_DELAY_POST_EXECUTION);
            
            logger.info("Initializing SensorHolder... done!");
        }

        logger.debug("initSensorHolder() - executed.");
    }


    /**
     * Initialize the <code>PredefinedSubscriptionHolder</code> searching for
     * predefined subscriptions.<br>
     * If properties "predefinedSubscriptionListPath" is not correctly defined
     * in the context the default path will be
     * "home.dir"/subscriptions/predefinedSubscriptionList.xml.<br>
     * Then the monitor is subscribed as a listener for predefined
     * subscription-related events.<br>
     * Finally the execution of the <code>PredefinedSubscriptionHolder</code>
     * is scheduled by setting the timer. The holder will be executed the first
     * time with a delay of four seconds, then repeatedly executed each minute.
     * 
     * @throws Exception
     *             This exception is thrown if the predefined subscription file
     *             does not contain the correct informations or is not formatted
     *             as expected.
     */
    protected void initPredefinedSubscriptionHolder() throws Exception {
        logger.debug("initPredefinedSubscriptionHolder() - executing method");
        
        if (predefinedSubscriptionHolder == null) {
            logger.info("Initializing PredefinedSubscriptionHolder...");
            
            predefinedSubscriptionHolder = new PredefinedSubscriptionHolder();
            predefinedSubscriptionHolder.addResourceListener(this);
            timer.schedule(predefinedSubscriptionHolder, 6000, 600000, TimerTask.EXECUTION_TYPE.FIXED_DELAY_POST_EXECUTION);
            
            logger.info("Initializing PredefinedSubscriptionHolder... done!");
        }

        logger.debug("initPredefinedSubscriptionHolder() - executed.");
    }




    /**
     * Initialize <code>TopicHolder</code> by just instantiating a new
     * TopicHolder class.
     * 
     * @throws Exception
     *             This exception is thrown if any error occurs while
     *             instantiating the TopicHolder class.
     */
    protected void initTopicHolder() throws Exception {
        logger.debug("initTopicHolder() - executing method");
        
        if (topicHolder == null) {
            logger.info("Initializing TopicHolder...");
            
            topicHolder = new TopicHolder(sensorHolder);
            
            logger.info("Initializing TopicHolder... done!");
        }

        logger.debug("initTopicHolder() - executed.");
    }


    /**
     * This method defines the things to do if a sensor causes an error. A
     * message containing the sensor name is logged and the sensor is removed
     * from the holder.
     * 
     * @param sensor
     *            The <code>Sensor</code> which caused an error.
     */
    public synchronized void doOnSensorError(Sensor sensor) throws RemoteException, GenericFault, AuthenticationFault, AuthorizationFault {
        logger.error("doOnSensorError() - SensorError!!! " + sensor.getName() + " => REMOVED !!!");
        removeSensor(sensor);
    }


    /**
     * This method defines the things to do when a sensor generates an event.<br>
     * First the correctness of the event is checked: the event should not be
     * null and its source sensor should be defined.<br>
     * Then this event is set as current event in the topic holder.
     * 
     * @param event
     *            The received <code>SensorEvent</code>.
     */
    public synchronized void doOnSensorEvent(SensorEvent event) {
        if (event == null) {
            return;
        }

        Sensor sensor = event.getSource();

        if (sensor != null) {
            try {
                if (topicHolder == null) {
                    logger.error("doOnSensorEvent() - topicHolder is NULL !!!");
                } else {
                    topicHolder.setCurrentEvent(sensor.getType(), event);
                }                
            } catch (IllegalArgumentException e) {
                logger.error("doOnSensorEvent() - Catched exception: [" + e.getMessage() + "]. Follow the stack trace:", e);
            } catch (TopicNotFoundException e) {
                logger.error("doOnSensorEvent() - Catched exception: [" + e.getMessage() + "]. Follow the stack trace:", e);
            }
        }

        timer.purge();
    }


    /**
     * Not yet implemented. Should be used for testing.
     * 
     * @throws RemoteException
     */
    public void ping() throws RemoteException, AuthenticationFault, AuthorizationFault {
        logger.debug("ping() - ping remote call invoked");
    }

    /**
     * Get the <code>Info</code> of this monitor. The informations passed are
     * retreived from the message context and are:<br>
     * 1) The monitor version.<br>
     * 2) A monitor description.<br>
     * 3) Available topics.<br>
     * 4) Available actions.
     * <p>
     * All these fields are filled with default values if not defined in the
     * context.
     * 
     * @return The required <code>Info</code> informations.
     * @throws RemoteException
     *             This method is exposed by the web service interface, so an
     *             exception is possible when remotely calling it.
     * @throws GenericFault
     *             This fault has a number of possible causes, for example a
     *             wrong interaction with topic holder or action holder.
     */
    public Info getInfo() throws RemoteException, GenericFault, AuthenticationFault, AuthorizationFault {
        logger.debug("getInfo() - getInfo remote call invoked. Returning the info object...");

        return info;
    }



    /**
     * Private method.
     * 
     * @param rate
     * @return
     */
    private NotificationHolder getNotificationHolder(SubscriptionPersistent subscription) throws IllegalArgumentException {
        if (subscription == null) {
            throw new IllegalArgumentException("subscription not specified!");
        }

        String name = null;

        if (notificationSharedThread) {
            name = "" + subscription.getPolicy().getRate();
        } else {
            name = subscription.getId();
        }
        
        NotificationHolder notificationHolder = null;

        if (notificationHolderList.containsKey(name)) {
            notificationHolder = (NotificationHolder) notificationHolderList.get(name);
        } else {
            notificationHolder = new NotificationHolder(name, subscription.getPolicy().getRate() * 1000, topicHolder);
            notificationHolder.setQueryProcessorHolder(queryProcessorHolder);
            notificationHolder.setActionHolder(actionHolder);
            notificationHolder.addResourceListener(this);

            notificationHolderList.put(name, notificationHolder);

           // timer.schedule(notificationHolder, 0, subscription.getPolicy().getRate() * 1000, TimerTask.EXECUTION_TYPE.FIXED_DELAY_POST_EXECUTION);
        }

        return notificationHolder;
    }


    /**
     * Private method.
     * 
     * @param subscription
     *            DOCUMENT ME!
     * @throws RemoteException
     *             DOCUMENT ME!
     * @throws SubscriptionFault
     *             DOCUMENT ME!
     * @throws GenericFault
     *             DOCUMENT ME!
     */
    private void checkSubscription(Subscription subscription) throws RemoteException, SubscriptionFault, GenericFault {
        logger.debug("checkSubscription() - checkSubscription remote call invoked");

        if (subscription == null) {
            throw (makeFault(SubscriptionFault.class, "checkSubscription", "0", "the subscription is null", "subscription == null"));
        }

        Calendar time = Calendar.getInstance();

        if (subscription.getExpirationTime() == null) {
            time.add(Calendar.HOUR_OF_DAY, maxExpirationTime_Hours);

            subscription.setExpirationTime(time);
        } else if (subscription.getExpirationTime().before(time)) {
            logger.warn("checkSubscription() - the subscription " + (subscription.getId()  != null? subscription.getId() : "") + " is expired (" + subscription.getExpirationTime().getTime() + ")");
            throw (makeFault(SubscriptionFault.class, "checkSubscription", "1", "The subscription expired on " + subscription.getExpirationTime().getTime(),
                    "The subscription is expired"));
        } else {
            time.add(Calendar.HOUR_OF_DAY, maxExpirationTime_Hours);

            if (time.before(subscription.getExpirationTime())) {
                subscription.setExpirationTime(time);
            }
        }

        if (subscription.getMonitorConsumerURL() == null) {
            logger.warn("checkSubscription() - MonitorConsumerURL not specified (subscription id=" + (subscription.getId()  != null? subscription.getId() : "N/A") + ")");
            throw (makeFault(SubscriptionFault.class, "checkSubscription", "2", "MonitorConsumerURL not specified", "MonitorConsumerURL not specified"));
        }

        Topic topic = subscription.getTopic();

        if (topic == null) {  
            logger.warn("checkSubscription() - Topic not specified (subscription id=" + (subscription.getId()  != null? subscription.getId() : "N/A") + ")");
            throw (makeFault(SubscriptionFault.class, "checkSubscription", "3", "Topic not specified", "topic == null"));
        }

        if (!topicHolder.checkSupportedTopic(topic.getName())) {
            logger.warn("checkSubscription() - Topic \"" + topic.getName() + "\" not supported (subscription id=" + (subscription.getId()  != null? subscription.getId() : "N/A") + ")");
            throw (makeFault(SubscriptionFault.class, "checkSubscription", "4", "Topic \"" + topic.getName() + "\" not supported",
                    "topicHolder.checkSupportedTopic( topic ) == false"));
        }

        org.glite.ce.monitorapij.types.Dialect[] dialects = topic.getDialect();
        Dialect dialect = null;
        org.glite.ce.monitorapij.resource.types.Topic systemTopic;
        try {
            systemTopic = topicHolder.getTopic(topic.getName());
            topic.setVisibility(systemTopic.getVisibility());
        } catch (IllegalArgumentException e) {
            logger.error("checkSubscription() - Catched IllegalArgumentException. Follow the Stack trace:", e);
            throw (makeFault(SubscriptionFault.class, "checkSubscription", "3", "Topic not specified", e.getMessage()));
        } catch (TopicNotFoundException e) {
            logger.error("checkSubscription() - Catched TopicNotFoundException. Follow the Stack trace:", e);
            throw (makeFault(SubscriptionFault.class, "checkSubscription", "4", "Topic \"" + topic.getName() + "\" not supported", e.getMessage()));
        }

        if ((dialects != null) && (dialects.length > 0)) {
            dialect = systemTopic.getDialect(dialects[0].getName());

            if (dialect == null) {
                logger.warn("checkSubscription() - Dialect \"" + dialects[0].getName() + "\" not supported (subscription id=" + (subscription.getId()  != null? subscription.getId() : "N/A") + ")");
                throw (makeFault(SubscriptionFault.class, "checkSubscription", "5", "Dialect \"" + dialects[0].getName() + "\" not supported", "isDialectSupported == false"));
            }
        } else {
            Dialect[] dialectArray = systemTopic.getDialect();

            if (dialectArray != null && (dialectArray.length > 0)) {
                topic.setDialect(new org.glite.ce.monitorapij.types.Dialect[] { new org.glite.ce.monitorapij.types.Dialect(dialectArray[0].getName(), dialectArray[0]
                        .getQueryLanguage()) });
            }
        }

        org.glite.ce.monitorapij.types.Policy policy = subscription.getPolicy();

        if (dialect != null) {
            if (policy.getRate() < 5) {
                policy.setRate(5);
            }
            org.glite.ce.monitorapij.types.Query query = policy.getQuery();

            if ((query != null) && (query.getQueryLanguage() != null)) {
                if (!dialect.checkSupportedQueryLanguage(query.getQueryLanguage())) {
                    logger.warn("checkSubscription() - QueryLanguage \"" + query.getQueryLanguage() + "\" not supported " + (subscription.getId()  != null? subscription.getId() : "N/A") + ")");
                    throw (makeFault(SubscriptionFault.class, "checkSubscription", "6", "QueryLanguage \"" + query.getQueryLanguage() + "\" not supported",
                            "dialect.checkSupportedQueryLanguage( query.getQueryLanguage() ) == false"));
                }
            }

            org.glite.ce.monitorapij.types.Action[] actionArray = policy.getAction();

            if (actionArray != null) {
                for (int i = 0; i < actionArray.length; i++) {
                    Action action = actionHolder.getAction(actionArray[i].getName());
                    if (action == null) {
                        logger.warn("checkSubscription() - Action \"" + actionArray[i].getName() + "\" not supported (subscription id=" + (subscription.getId()  != null? subscription.getId() : "N/A") + ")");
                        throw (makeFault(SubscriptionFault.class, "checkSubscription", "6", "Action \"" + actionArray[i].getName() + "\" not supported", ""));
                    }
                }
            } 
        }

        if (policy != null) {
            if (policy.getRate() < 5) {
                policy.setRate(5);
            }
        } else {
            policy = new org.glite.ce.monitorapij.types.Policy();
            policy.setRate(5);
        }
        subscription.setPolicy(policy);
    }

    private String getUserDN_FQAN(String vo) throws GenericFault {
        String fqan = "";
        List<String> fqanlist = null;

        if (vo != null) {
            fqanlist = CEUtils.getFQAN(vo);
        } else {
            String userVO = CEUtils.getUserDefaultVO();

            if (userVO != null) {
                fqanlist = CEUtils.getFQAN(userVO);
            }
        }

        if (fqanlist != null && fqanlist.size() > 0) {
            fqan = normalize(fqanlist.get(0).toString());
        }

        return normalize(CEUtils.getUserDN_X500()) + fqan;
    }
    
    private String normalize(String s) {
        if (s != null) {
            String toReturn = s.replaceAll("\\W", "_");

            logger.debug("normalize() - Returning normalized string [" + toReturn + "]");

            return toReturn;
        }

        logger.debug("normalize() - Returning empty string.");

        return null;
    }


    /**
     * Perform the specified subscription setting it as persistent.
     * 
     * @param subscription
     *            The <code>Subscription</code> to be performed.
     * @return The <code>SubscriptionRef</code> relative to the subscription.
     * @throws RemoteException
     *             Thrown when an error occurs during a remote call of this
     *             method.
     * @throws SubscriptionFault
     *             This exception is thrown if the subscription is not correctly
     *             specified.
     * @throws GenericFault
     *             A number of other possible errors.
     */
    public SubscriptionRef subscribe(Subscription subscription) throws RemoteException, SubscriptionFault, GenericFault, AuthenticationFault, AuthorizationFault {
        logger.debug("subscribe() - subscribe remote call invoked");

        return subscribe(subscription, true);
    }


    /**
     * Perform the specified subscription setting the persistent flag as
     * specified.
     * 
     * @param subscription
     *            The subscription to be performed.
     * @param persistent
     *            This flag is used to set the persistance of the subscription.
     *            If a subscription if persistent it is added to the
     *            subscription registry.
     * @return The <code>SubscriptionRef</code> relative to the subscription.
     * @throws RemoteException
     *             Thrown when an error occurs during a remote call of this
     *             method.
     * @throws SubscriptionFault
     *             This exception is thrown if the subscription is not correctly
     *             specified.
     * @throws GenericFault
     *             A number of other possible errors.
     */
    public SubscriptionRef subscribe(Subscription subscription, boolean persistent) throws RemoteException, SubscriptionFault, GenericFault, AuthenticationFault, AuthorizationFault {
        logger.debug("subscribe() - subscribe remote call invoked");

        checkSubscription(subscription);

        try {
            SubscriptionPersistent subscriptionPersistent = subscriptionRegistry.makeSubscriptionPersistent(subscription);

            if( (subscriptionPersistent.getPolicy().getAction() == null) || (subscriptionPersistent.getPolicy().getAction().length == 0) ) {
                logger.warn("subscribe() - subscriptionPersistent.getPolicy().getAction() is NULL. Creating it...");

                Action[] acts = new Action[1];
                acts[0] = new Action("SendNotification");
                acts[0].setDoActionWhenQueryIs(true);
                subscriptionPersistent.getPolicy().setAction(acts);
            }

            String voName = CEUtils.getUserDefaultVO();
            if (voName != null) {
                subscriptionPersistent.setSubscriberId(getUserDN_FQAN(voName));
                subscriptionPersistent.setSubscriberGroup(voName);
            } else {
                subscriptionPersistent.setSubscriberId(getUserDN_FQAN(null));
            }

            addToNotificationHolder(subscriptionPersistent);
            subscriptionRegistry.register(subscriptionPersistent);

            SubscriptionRef subscriptionRef = new SubscriptionRef();
            subscriptionRef.setSubscriptionID(subscriptionPersistent.getId());
            subscriptionRef.setExpirationTime(subscriptionPersistent.getExpirationTime());

            logger.info("subscribe() - new subscription registered: subscription id = " + subscriptionPersistent.getId() + " - subscriber id = "
                    + subscriptionPersistent.getSubscriberId());

            return subscriptionRef;
        } catch (Exception e) {
            logger.error("subscribe() - Catched exception: [" + e.getMessage() + "]. Follow the Stack trace:", e);

            BaseFaultType fault = makeFault(GenericFault.class, "subscribe", "*", "internal exception", e.getMessage());

            throw (fault);
        }
    }


    /**
     * Private method.
     * 
     * @param subscription
     *            DOCUMENT ME!
     */
    protected void addToNotificationHolder(SubscriptionPersistent subscription) {
        org.glite.ce.monitorapij.resource.types.Policy policy = subscription.getPolicy();

        int rate = 0;

        if (policy != null) {
            rate = policy.getRate();
            // rate lower than 5 sec.
            rate = (rate < 5) ? 5 : rate;
        } else {
            subscription.setPolicy(new org.glite.ce.monitorapij.resource.types.Policy());

            Sensor sensor = sensorHolder.getSensorByType(subscription.getTopic().getName());
            rate = Integer.parseInt(((Resource) sensor).getProperty("executionDelay").getValue());
        }
        
        subscription.getPolicy().setRate(rate);

        NotificationHolder notificationHolder = getNotificationHolder(subscription);

        logger.debug("addToNotificationHolder() - Adding SUBSCRIPTION to NotificationHolder: " + subscription.getCompleteInfo());

        notificationHolder.addSubscription(subscription);
    }



    /**
     * Unsubscribe the subscription specified by the relative
     * <code>SubscriptionRef</code>.
     * 
     * @param subscriptionRef
     *            The <code>SubscriptionRef</code> relative to the
     *            subscription.
     * @throws RemoteException
     *             Thrown when an error occurs during a remote call of this
     *             method.
     * @throws GenericFault
     *             A number of other possible errors.
     */
    public void unsubscribe(SubscriptionRef subscriptionRef) throws RemoteException, GenericFault, AuthenticationFault, AuthorizationFault, SubscriptionNotFoundFault {
        logger.debug("unsubscribe() - unsubscribe remote call invoked");

        if (subscriptionRef == null) {
            logger.error("unsubscribe() - subscriptionRef argument is NULL!");

            throw (makeFault(GenericFault.class, "unsubscribe", "0", "the subscriptionRef is null", "subscriptionRef == null"));
        }

        String subscriptionID = subscriptionRef.getSubscriptionID();

        if ((subscriptionID == null) || subscriptionID.equals("")) {
            logger.error("unsubscribe() - SubscriptionID not specified");

            throw (makeFault(GenericFault.class, "unsubscribe", "0", "subscriptionID not specified", "subscriptionID == null || subscriptionID.equals( \"\" )"));
        }

        try {
            SubscriptionPersistent subscription = subscriptionRegistry.getSubscription(subscriptionID);
            NotificationHolder notificationHolder = getNotificationHolder(subscription);
            notificationHolder.removeSubscription(subscription);

            if (notificationHolder.getSubscriptionListSize() == 0) {
                notificationHolder.destroy();
                notificationHolderList.remove(notificationHolder.getName());                
            }

            logger.debug("unsubscribe() - Unregistering Subscription [" + subscriptionID + "]");

            subscriptionRegistry.unregister(subscriptionID);
        } catch (Exception e) {
            logger.error("unsubscribe() - Subscription [" + subscriptionID + "] NOT FOUND! Follow the Stack trace:", e);

            throw (makeFault(SubscriptionNotFoundFault.class, "unsubscribe", "0", "subscription \"" + subscriptionID + "\" not found!", "subscription \"" + subscriptionID
                    + "\" not found!"));
        }
    }

    public SubscriptionRef update(Subscription subscription) throws RemoteException, SubscriptionFault, GenericFault, AuthenticationFault, AuthorizationFault {
        logger.debug("update() - update remote call invoked");

        checkSubscription(subscription);
        SubscriptionPersistent subscriptionPersistent;
        try {
            subscriptionPersistent = subscriptionRegistry.getSubscription(subscription.getId());
        } catch (NamingException e) {
            logger.error("unsubscribe() - NamingException catched. Follow the Stack trace:", e);
            throw (makeFault(SubscriptionFault.class, "update", "0", "Subscription \"" + subscription.getId() + "\" not found!", e.getMessage()));
        }

        if (subscriptionPersistent == null) {
            String msg = "subscription [" + subscription.getId() + "] not found!";

            logger.error("update() - " + msg);
            
            throw makeFault(SubscriptionFault.class, "update", "0", msg, msg);
        }

        if (!subscriptionPersistent.getSubscriberId().equals(getUserDN_FQAN(null))) {
            String msg = "subscription update not allowed: the subscription [" + subscription.getId() + "] belongs to another client!";
            logger.error("update() - " + msg);

            throw makeFault(AuthorizationFault.class, "update", "1", msg, msg);
        }

        NotificationHolder notificationHolder = getNotificationHolder(subscriptionPersistent);
        notificationHolder.removeSubscription(subscriptionPersistent);

        try {
            SubscriptionPersistent newSubscriptionPersistent = subscriptionRegistry.makeSubscriptionPersistent(subscription);

            subscriptionPersistent.setTopic(newSubscriptionPersistent.getTopic());
            subscriptionPersistent.setPolicy(newSubscriptionPersistent.getPolicy());
            subscriptionPersistent.setMonitorConsumerURL(newSubscriptionPersistent.getMonitorConsumerURL());
            subscriptionPersistent.setExpirationTime(subscription.getExpirationTime());
            subscriptionPersistent.setType(subscription.getType());

            if( (subscriptionPersistent.getPolicy().getAction() == null) || (subscriptionPersistent.getPolicy().getAction().length == 0) ) {
                logger.warn("update() - subscriptionPersistent.getPolicy().getAction() is NULL. Creating it...");

                Action[] acts = new Action[1];
                acts[0] = new Action("SendNotification");
                acts[0].setDoActionWhenQueryIs(true);
                subscriptionPersistent.getPolicy().setAction(acts);
            }
            
            if (subscription.getJarPath() != null) {
                subscriptionPersistent.setJarPath(new URI(subscription.getJarPath().toString()));
            }

            org.glite.ce.monitorapij.types.Property[] property = subscription.getProperty();
            if (property != null) {
                Property[] newProperty = new Property[property.length];

                for (int i = 0; i < property.length; i++) {
                    newProperty[i] = new Property(property[i].getName(), property[i].getValue());
                }
                subscriptionPersistent.setProperty(newProperty);
            }
            
            notificationHolder.addSubscription(subscriptionPersistent);
                        
            subscriptionRegistry.update(subscriptionPersistent);

            SubscriptionRef subscriptionRef = new SubscriptionRef();
            subscriptionRef.setSubscriptionID(subscriptionPersistent.getId());
            subscriptionRef.setExpirationTime(subscriptionPersistent.getExpirationTime());

            logger.info("update() - Subscription [" + subscription.getId() + "] updated!");

            return subscriptionRef;
        } catch (Exception e) {
            logger.error("update() - Catched Exception: [" + e.getMessage() + "]. Follow the Stack trace:", e);

            BaseFaultType fault = makeFault(GenericFault.class, "subscribe", "*", "internal exception", e.getMessage());

            logger.error("update() - " + fault);
            
            throw (fault);
        }
    }


    /**
     * Stop the notification sending relative to the subscription specified by
     * the relative <code>SubscriptionRef</code>.<br>
     * This is done by calling the <code>SubscriptionPersistent</code>.pause()
     * method.
     * 
     * @param subscriptionRef
     *            The <code>SubscriptionRef</code> relative to the
     *            subscription.
     * @throws RemoteException
     *             Thrown when an error occurs during a remote call of this
     *             method.
     * @throws GenericFault
     *             Thrown when the subscriptionRef is null, or it does not
     *             specify the subsciptionID, or the subscription is not present
     *             in the registry.
     */
    public void pauseSubscription(SubscriptionRef subscriptionRef) throws RemoteException, GenericFault, AuthenticationFault, AuthorizationFault, SubscriptionNotFoundFault {
        logger.debug("pauseSubscription() - pauseSubscription remote call invoked");

        if (subscriptionRef == null) {
            logger.error("pauseSubscription() - subscriptionRef argument is NULL!");

            throw (makeFault(GenericFault.class, "pauseSubscription", "0", "the subscriptionRef is null", "subscriptionRef == null"));
        }

        String subscriptionID = subscriptionRef.getSubscriptionID();

        if ((subscriptionID == null) || subscriptionID.equals("")) {
            logger.error("pauseSubscription() - Subscription ID is null or an empty string...");

            throw (makeFault(GenericFault.class, "pauseSubscription", "0", "subscriptionID not specified", "subscriptionID == null || subscriptionID.equals( \"\" )"));
        }

        try {
            SubscriptionPersistent subscription = subscriptionRegistry.getSubscription(subscriptionID);
            subscription.pause();
            subscriptionRegistry.update(subscription); // Save the modified subscription into the persistent store
            logger.info("pauseSubscription() - Subscription [" + subscription.getId() + "] suspended!");
        } catch (Exception e) {
            logger.error("pauseSubscription() - Catched Exception: [" + e.getMessage() + "]. Follow the Stack trace:", e);

            throw (makeFault(SubscriptionNotFoundFault.class, "pauseSubscription", "0", "subscription \"" + subscriptionID + "\" not found!", "subscription \"" + subscriptionID
                    + "\" not found!"));
        }
    }



    /**
     * Resume the notification sending relative to the subscription specified by
     * the relative <code>SubscriptionRef</code>.<br>
     * This is done by calling the <code>SubscriptionPersistent</code>.resume()
     * method.
     * 
     * @param subscriptionRef
     *            The <code>SubscriptionRef</code> relative to the
     *            subscription.
     * @throws RemoteException
     *             Thrown when an error occurs during a remote call of this
     *             method.
     * @throws GenericFault
     *             Thrown when the subscriptionRef is null, or it does not
     *             specify the subsciptionID, or the subscription is not present
     *             in the registry.
     */
    public void resumeSubscription(SubscriptionRef subscriptionRef) throws RemoteException, GenericFault, AuthenticationFault, AuthorizationFault, SubscriptionNotFoundFault {
        logger.debug("resumeSubscription() - resumeSubscription remote call invoked");

        if (subscriptionRef == null) {
            logger.error("resumeSubscription() - subscriptionRef argument is NULL!");

            throw (makeFault(GenericFault.class, "resumeSubscription", "0", "the subscriptionRef is null", "subscriptionRef == null"));
        }

        String subscriptionID = subscriptionRef.getSubscriptionID();

        if ((subscriptionID == null) || subscriptionID.equals("")) {
            logger.error("resumeSubscription() - Subscription ID is null or an empty string...");

            throw (makeFault(GenericFault.class, "resumeSubscription", "0", "subscriptionID not specified", "subscriptionID == null || subscriptionID.equals( \"\" )"));
        }

        try {
            SubscriptionPersistent subscription = subscriptionRegistry.getSubscription(subscriptionID);
            subscription.resume();
            subscriptionRegistry.update(subscription);// Save the modified subscription into the persistent store
            logger.info("resumeSubscription() - Subscription [" + subscription.getId() + "] resumed!");
        } catch (Exception e) {
            logger.error("resumeSubscription() - Catched Exception: [" + e.getMessage() + "]. Follow the Stack trace:", e);
            throw (makeFault(SubscriptionNotFoundFault.class, "resumeSubscription", "0", "subscription \"" + subscriptionID + "\" not found!", "subscription \"" + subscriptionID
                    + "\" not found!"));
        }
    }



    /**
     * Get the array of <code>Topic</code>s registered on this monitor.
     * 
     * @return The array of <code>Topic</code>s registered on this monitor.
     * @throws RemoteException
     *             Thrown when an error occurs during a remote call of this
     *             method.
     * @throws GenericFault
     *             A number of other possible errors.
     */
    public Topic[] getTopics() throws RemoteException, GenericFault, AuthenticationFault, AuthorizationFault {
        logger.debug("getTopics() - getTopics remote call invoked");

        if (topicHolder != null) {
            org.glite.ce.monitorapij.resource.types.Topic[] topicArray = topicHolder.getTopics();
            org.glite.ce.monitorapij.types.Topic[] newTopic = new org.glite.ce.monitorapij.types.Topic[topicArray.length];

            for (int i = 0; i < topicArray.length; i++) {
                org.glite.ce.monitorapij.resource.types.Topic topic = topicArray[i];
                Dialect[] dialect = topic.getDialect();

                org.glite.ce.monitorapij.types.Dialect[] newDialect = null;
                if (dialect != null) {
                    newDialect = new org.glite.ce.monitorapij.types.Dialect[dialect.length];

                    for (int j = 0; j < dialect.length; j++) {
                        newDialect[j] = new org.glite.ce.monitorapij.types.Dialect(dialect[j].getName(), dialect[j].getQueryLanguage());
                    }
                }

                newTopic[i] = new org.glite.ce.monitorapij.types.Topic(topic.getName(), newDialect, topic.getVisibility());
            }

            if(logger.isDebugEnabled()) {
                for (int i = 0; i < newTopic.length; i++) {
                    logger.debug("getTopics() - Returning topic [" + newTopic[i].getName() + "]");
                }
            }

            return newTopic;
        }

        return new Topic[0];
    }

    /**
     * Get the last event relative to the specified topic.
     * 
     * @param topic
     *            The specified <code>Topic</code>.<br>
     *            This must be supported by the topic holder. In particular, the
     *            name and the supported dialect will be compared.
     * @return The searched event. It will be formatted using the dialect
     *         specified by the topic parameter.
     * @throws TopicNotSupportedFault
     *             If specified topic is not supported by the
     *             <code>TopicHolder</code>.
     * @throws DialectNotSupportedFault
     *             The dialect specified by the topic parameter is not supported
     *             by the topic saved in the topic holder.
     * @throws RemoteException
     *             Thrown when an error occurs during a remote call of this
     *             method.
     * @throws GenericFault
     *             A number of other possible errors.
     */
    public Event getEvent(Topic topic) throws RemoteException, TopicNotSupportedFault, DialectNotSupportedFault, GenericFault, AuthenticationFault, AuthorizationFault {
        logger.debug("getEvent() - getEvent remote call invoked");

        Event[] event = getTopicEvent(topic);
        if (event != null && event.length > 0) {
            return event[0];
        }

        return null;
    }


    public Event[] getTopicEvent(Topic topic) throws RemoteException, AuthenticationFault, GenericFault, DialectNotSupportedFault, AuthorizationFault, TopicNotSupportedFault {
        logger.debug("getTopicEvent() - getTopicEvent remote call invoked");

        if (topic == null) {
            throw (makeFault(GenericFault.class, "getEvent", "0", "the topicExpression is null", "topic == null"));
        }

        org.glite.ce.monitorapij.resource.types.Topic systemTopic;
        try {
            systemTopic = topicHolder.getTopic(topic.getName());
        } catch (IllegalArgumentException e) {
            logger.error("getTopicEvent() - IllegalArgumentException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);

            throw (makeFault(TopicNotSupportedFault.class, "getTopicEvent", "2", "Topic not specified", e.getMessage()));

        } catch (TopicNotFoundException e) {
            logger.error("getTopicEvent() - TopicNotFoundException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);

            throw (makeFault(TopicNotSupportedFault.class, "getTopicEvent", "3", "Topic \"" + topic.getName() + "\" not supported", e.getMessage()));

        }

        org.glite.ce.monitorapij.types.Dialect[] dialect = topic.getDialect();

        if ((dialect != null) && (dialect.length > 0)) {
            if (!systemTopic.checkSupportedDialect(dialect[0].getName())) {
                throw (makeFault(DialectNotSupportedFault.class, "getEvent", "4", "Dialect \"" + dialect[0].getName() + "\" not supported!", "isDialectSupported == false"));
            }
        } else {
            dialect = null;
        }

        String userVO = CEUtils.getUserDefaultVO();
        String subscriberId = getUserDN_FQAN(userVO);

        SensorEventArrayList eventList;
        try {
            eventList = topicHolder.getEvents(topic.getName(), subscriberId, userVO);
        } catch (IllegalArgumentException e) {
            logger.error("getTopicEvent() - IllegalArgumentException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
            throw (makeFault(SubscriptionFault.class, "getTopicEvent", "5", "Topic not specified", e.getMessage()));
        } catch (TopicNotFoundException e) {
            logger.error("getTopicEvent() - TopicNotFoundException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
            throw (makeFault(SubscriptionFault.class, "getTopicEvent", "6", "Topic \"" + topic.getName() + "\" not supported", e.getMessage()));
        }

        if (dialect != null && dialect.length > 0) {
            String dataFormat = dialect[0].getName();

            for (int i = 0; i < eventList.size(); i++) {
                SensorEvent event = (SensorEvent) eventList.get(i);

                SensorOutputDataFormat format = event.getSensorOutputDataFormatApplied();
                if (format == null || !format.getName().equals(dataFormat)) {
                    try {
                        event.applyFormat(dataFormat);
                    } catch (SensorException e) {
                        logger.error("getTopicEvent() - SensorException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
                        throw (makeFault(GenericFault.class, "getEvent", "0", e.getMessage(), ""));
                    }
                }
            }
        }

        Event[] topicEvent = new Event[eventList.size()];
        for (int i = 0; i < eventList.size(); i++) {
            SensorEvent event = (SensorEvent) eventList.get(i);
            topicEvent[i] = new Event(event.getID(), event.getTimestamp(), event.getMessage(), event.getProducer());

        }

        return topicEvent;
    }


    /**
     * Log a notification message with some informations about the source and
     * the message of the event.<br>
     * The event is catched from the notification using the
     * <code>Notification</code>.getEvent() method.
     * 
     * @param notification
     *            The specified <code>Notification</code>.
     * @throws RemoteException
     *             Thrown when an error occurs during a remote call of this
     *             method.
     */
    public void notify(Notification notification) throws RemoteException {
//        String msg = "\n-------------------------------------------------------------------------";
//        msg += ("\n- consumer received a new notification produced by " + notification.getEvent().getProducer());
//        msg += "\n-------------------------------------------------------------------------";
//
//        String[] message = notification.getEvent().getMessage();
//
//        for (int i = 0; i < message.length; i++) {
//            msg += "\n**********************************************";
//            msg += ("\n* notification message number: " + i);
//            msg += "\n**********************************************\n";
//            msg += (message[i] + "\n\n");
//        }
//
//        msg += "\n-------------------------------------------------------------------------";
//
//        LogUtils.logDebug(msg);
    }



    /**
     * Private method.
     * 
     * @param faultclass
     *            DOCUMENT ME!
     * @param methodName
     *            DOCUMENT ME!
     * @param errorCode
     *            DOCUMENT ME!
     * @param description
     *            DOCUMENT ME!
     * @param faultCause
     *            DOCUMENT ME!
     * @return DOCUMENT ME!
     */
    private BaseFaultType makeFault(Class faultclass, String methodName, String errorCode, String description, String faultCause) {
        BaseFaultType fault = new BaseFaultType();

        try {
            fault = (BaseFaultType) faultclass.newInstance();
            fault.setMethodName(methodName);
            fault.setDescription(description);
            fault.setErrorCode(errorCode);
            fault.setFaultCause(faultCause);
            fault.setTimestamp(Calendar.getInstance());
        } catch (InstantiationException e) {
            logger.error("makeFault() - InstantiationException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
        } catch (IllegalAccessException e) {
            logger.error("makeFault() - IllegalAccessException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
        }

        return fault;
    }

    /**
     * This method performs the operations to do when the specified
     * <code>ResourceEvent</code> occurs.<br>
     * First the holder relative to the specified event is checked to verify if
     * it is an instance of <code>SubscriptionPersistent</code> or
     * <code>QueryProcessorHolder</code> or <code>ActionHolder</code> or
     * <code>SensorHolder</code>.
     * <p>
     * In the first case the resources relative to this event are subscriptions.<br>
     * The possible types of event are specified in the constant fields of
     * <code>ResourceEvent</code>.<br>
     * If so some actions are performed depending on the type of event:<br>
     * 1) RESOURCE_ADDED: the subscription is checked and then added to the
     * notification holder.<br>
     * 2) RESOURCE_UPDATED: the subscription is removed from the notification
     * holder and then re-added with updated value.<br>
     * 3) RESOURCE_REMOVED: the subscription is removed from the notification
     * holder.
     * <p>
     * If the specified event is an instance of <code>SensorHolder</code> then
     * the resources relative to this event are subscriptions. The specified
     * sensors are added or removed or updated depending on event type.
     * <p>
     * If the specified event is an instance of
     * <code>QueryProcessorHolder</code> or <code>ActionHolder</code>, a
     * message is just logged specifying the source and the nature of the event.
     * <p>
     * 
     * @param event
     *            The specified <code>ResourceEvent</code>.
     */
    public synchronized void doOnResourceEvent(ResourceEvent event) {
        if (event == null) {
            return;
        }

        ResourceHolder holder = event.getResourceHolder();
        List<Resource> resourceList = event.getResources();

        logger.debug("doOnResourceEvent() - " + holder.getName() + " resources=" + resourceList.size() + " eventId=" + event.getEvent());

        for (Resource resource : resourceList) {
            switch (event.getEvent()) {
            case ResourceEvent.RESOURCE_ADDED:
                if (resource instanceof SubscriptionPersistent) {
                    SubscriptionPersistent subscription = (SubscriptionPersistent) resource;
                    try {
                        checkSubscription(subscriptionRegistry.makeSubscription(subscription));
                    } catch (Throwable e) {
                        logger.error("doOnResourceEvent() - Throwable catched [" + e.getMessage() + "]. Follow the Stack trace:", e);
                        break;
                    }

                    subscription.setRemoveOnStartup(true);
                    try {
                        subscriptionRegistry.register(subscription);
                    } catch (Throwable e) {
                        logger.error("doOnResourceEvent() - Throwable catched [" + e.getMessage() + "]. Follow the Stack trace:", e);
                        break;
                    }
                    
                    addToNotificationHolder(subscription);

                    logger.info("doOnResourceEvent() - added new subscription [id = " + subscription.getId() + "]");                    
                } else if (resource instanceof Sensor) {
                    try {
                        addSensor((Sensor)resource);
                        logger.info("doOnResourceEvent() - added new sensor [name=" + resource.getName() + "]");
                    } catch (SensorException e) {
                        logger.error("doOnResourceEvent() - SensorException catched [" + e.getMessage() + "]. Follow the Stack trace:", e);
                    } catch (Throwable e) {
                        logger.error("doOnResourceEvent() - Throwable catched [" + e.getMessage() + "]. Follow the Stack trace:", e);
                    }
                } else if (resource instanceof QueryProcessor) {
                    logger.info("doOnResourceEvent() - added new queryProcessor [name=" + resource.getName() + "]");
                } else if (resource instanceof Action) {
                    logger.info("doOnResourceEvent() - added new action [name=" + resource.getName() + "]");
                }
                
                break;

            case ResourceEvent.RESOURCE_UPDATED:
                if (resource instanceof SubscriptionPersistent) {
                    SubscriptionPersistent subscription = (SubscriptionPersistent) resource;
                    subscription.setRemoveOnStartup(true);

                    try {
                        subscriptionRegistry.update(subscription);
                    } catch (Exception e) {
                        logger.error("doOnResourceEvent() - Exception catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
                        try {
                            subscriptionRegistry.register(subscription);
                        } catch (Throwable t) {
                            logger.error("doOnResourceEvent() - Throwable catched [" + e.getMessage() + "]. Follow the Stack trace:", t);
                            break;
                        }
                    }

                    NotificationHolder notificationHolder = (NotificationHolder)holder;
                    notificationHolder.removeSubscription(subscription);
                    notificationHolder.addSubscription(subscription);

                    logger.info("doOnResourceEvent() - updated subscription [id = " + subscription.getId() + "]");    
                } else if (resource instanceof Sensor) {
                    try {
                        removeSensor((Sensor)resource);
                        addSensor(sensorHolder.getSensorById(resource.getId()));
                        logger.info("doOnResourceEvent() - updated sensor [name=" + resource.getName() + "]");
                    } catch (SensorException e) {
                        logger.error("doOnResourceEvent() - SensorException catched [" + e.getMessage() + "]. Follow the Stack trace:", e);
                    } catch (Throwable e) {
                        logger.error("doOnResourceEvent() - Throwable catched [" + e.getMessage() + "]. Follow the Stack trace:", e);
                    }
                } else if (resource instanceof QueryProcessor) {
                    logger.info("doOnResourceEvent() - updated queryProcessor [" + resource.getName() + "]");
                } else if (resource instanceof Action) {
                    logger.info("doOnResourceEvent() - updated action [" + resource.getName() + "]");
                }

                break;
                
            case ResourceEvent.RESOURCE_REMOVED:
                if (resource instanceof SubscriptionPersistent) {
                    SubscriptionPersistent subscription = (SubscriptionPersistent) resource;
                    NotificationHolder notificationHolder = (NotificationHolder)holder;
                    notificationHolder.removeSubscription(subscription);

                    try {
                        subscriptionRegistry.unregister(subscription);
                    } catch (Throwable e) {
                        logger.error("doOnResourceEvent() - Throwable catched [" + e.getMessage() + "]. Follow the Stack trace:", e);
                        break;
                    }

                    if (notificationHolder.getSubscriptionListSize() == 0) {
                        notificationHolder.destroy();
                        notificationHolderList.remove(notificationHolder.getName());
                        logger.info("doOnResourceEvent() - removed notificationHolder [name = " + notificationHolder.getName() + "]");
                    }

                    logger.info("doOnResourceEvent() - unregistered subscription [id = " + subscription.getId() + "]");
                } else if (resource instanceof Sensor) {
                    try {
                        removeSensor((Sensor)resource);
                        logger.info("doOnResourceEvent() - removed sensor [" + resource.getName() + "]");
                    } catch (Throwable e) {
                        logger.error("doOnResourceEvent() - Throwable catched [" + e.getMessage() + "]. Follow the Stack trace:", e);
                    }
                } else if (resource instanceof QueryProcessor) {
                    logger.info("doOnResourceEvent() - removed queryProcessor [" + resource.getName() + "]");
                } else if (resource instanceof Action) {
                    logger.info("doOnResourceEvent() - removed action [" + resource.getName() + "]");
                }
                
                break;
            }
        }
    }



    /**
     * Private method.
     * 
     * @param sensor
     *            DOCUMENT ME!
     * @throws SensorException
     *             DOCUMENT ME!
     */
    private void addSensor(Sensor sensor) throws RemoteException, GenericFault, AuthenticationFault, AuthorizationFault, SensorException {
        if (sensor == null) {
            return;
        }

        logger.debug("addSensor() - Checking if new sensors have been added... " + sensor.getName());

        String sensorType = sensor.getType();

        logger.debug("addSensor() - SensorType is null " + (sensorType==null));
        
        if (sensorType != null) {
            sensor.init();
            sensor.addSensorListener(this);

            SensorOutputDataFormat[] formats = sensor.getFormats();
            Dialect[] dialects = new Dialect[formats.length];

            for (int i = 0; i < formats.length; i++) {
                dialects[i] = new Dialect(formats[i].getName());
                dialects[i].setQueryLanguage(formats[i].getSupportedQueryLang());
            }

            org.glite.ce.monitorapij.resource.types.Topic topic = new org.glite.ce.monitorapij.resource.types.Topic(sensorType);
            topic.setPurgeAllEventsOnStartup( sensor.isPurgeAllEventsOnStartup() );
            topic.setDialect(dialects);
            topic.setEventOverwriteModeActive(sensor.isEventOverwriteModeActive());
            topic.setVisibility((String) visibilityTable.get(sensor.getScope()));

            try {
                topicHolder.addTopic(topic);
            } catch (IllegalArgumentException e) {
                logger.error("addSensor() - IllegalArgumentException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
            } catch (NameAlreadyBoundException e) {
                logger.error("addSensor() - NameAlreadyBoundException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
            } catch (NamingException e) {
                logger.error("addSensor() - NamingException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
            }

            try {
                sensor.startSensor();
            } catch (Exception e) {
                // e1.printStackTrace();
                logger.error("addSensor() - Exception catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
                removeSensor(sensor);
            }
        }

        try {
            this.updateInfo();
        } catch (Throwable e) {
            logger.error("addSensor() - Throwable catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
        }
    }


    /**
     * Private method.
     * 
     * @param sensor
     *            DOCUMENT ME!
     */
    private void removeSensor(Sensor sensor) throws RemoteException, GenericFault, AuthenticationFault, AuthorizationFault {
        if (sensor == null) {
            return;
        }

        logger.debug("addSensor() - Checking if some sensor have been removed...");

        sensor.removeSensorListener(this);

        try {
            topicHolder.removeTopic(sensor.getType());
        } catch (IllegalArgumentException e) {
            logger.error("removeSensor() - IllegalArgumentException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
        } catch (TopicNotFoundException e) {
            logger.error("removeSensor() - TopicNotFoundException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
        }
        sensor.destroySensor();

        this.updateInfo();
    }

    public SubscriptionRef[] getSubscriptionRef() throws RemoteException, AuthenticationFault, GenericFault, AuthorizationFault {
        logger.debug("getSubscriptionRef() - getSubscriptionRef remote call invoked.");

        ArrayList<SubscriptionPersistent> subscriptions = subscriptionRegistry.getSubscriptionBySubscriberId(getUserDN_FQAN(null));
        SubscriptionRef[] ref = null;
        if (subscriptions != null) {
            ref = new SubscriptionRef[subscriptions.size()];

            for (int i = 0; i < ref.length; i++) {
                SubscriptionPersistent subscription = (SubscriptionPersistent) subscriptions.get(i);

                ref[i] = new SubscriptionRef();
                ref[i].setSubscriptionID(subscription.getId());
                ref[i].setExpirationTime(subscription.getExpirationTime());

                logger.debug("getSubscriptionRef() - Added to response subscription [" + subscription.getCompleteInfo() + "]");
            }
        } else {
            ref = new SubscriptionRef[0];
        }

        return ref;
    }
    

    public Subscription[] getSubscription(SubscriptionRef[] refList) throws RemoteException, AuthenticationFault, GenericFault, AuthorizationFault {
        logger.debug("getSubscription() - getSubscription remote call invoked.");

        ArrayList<SubscriptionPersistent> subscriptions = null;
        String subscriberId = getUserDN_FQAN(null);

        if (refList == null || refList.length == 0) {
            subscriptions = subscriptionRegistry.getSubscriptionBySubscriberId(subscriberId);
        } else {
            subscriptions = new ArrayList<SubscriptionPersistent>(0);
            for (int i = 0; i < refList.length; i++) {
                SubscriptionPersistent subscription;
                try {
                    subscription = subscriptionRegistry.getSubscription(refList[i].getSubscriptionID());
                } catch (NamingException e) {
                    logger.error("getSubscription() - NamingException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
                    throw (makeFault(GenericFault.class, "getSubscription", "0", e.getMessage(), e.getMessage()));
                }

                if (subscription.getSubscriberId().equals(subscriberId)) {

                    logger.debug("getSubscription() - Adding subscription [" + subscription.getCompleteInfo() + "]");

                    subscriptions.add(subscription);
                }
            }
        }

        if (subscriptions == null || subscriptions.size() == 0) {
            return new Subscription[0];
        }

        Subscription[] subscriptionArray = new Subscription[subscriptions.size()];

        for (int i = 0; i < subscriptions.size(); i++) {
            try {
                subscriptionArray[i] = subscriptionRegistry.makeSubscription((SubscriptionPersistent) subscriptions.get(i));
            } catch (IllegalArgumentException e) {
                logger.error("getSubscription() - IllegalArgumentException catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
                throw (makeFault(GenericFault.class, "getSubscription", "1", e.getMessage(), e.getMessage()));
            } catch (Exception e) {
                // e.printStackTrace();
                logger.error("getSubscription() - Exception catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
                throw (makeFault(GenericFault.class, "getSubscription", "2", e.getMessage(), e.getMessage()));
            }
        }

        return subscriptionArray;
    }
    
    protected void updateInfo() throws RemoteException, GenericFault, AuthenticationFault, AuthorizationFault {
        synchronized (info) {
            logger.debug("updateInfo() - Updating service's information...");

            if (actionHolder != null) {
                Action[] action = actionHolder.getActions();

                org.glite.ce.monitorapij.types.Action[] newAction = new org.glite.ce.monitorapij.types.Action[action.length];

                for (int i = 0; i < action.length; i++) {
                    newAction[i] = new org.glite.ce.monitorapij.types.Action();
                    newAction[i].setName(action[i].getName());

                    logger.debug("updateInfo() - Added Action [" + action[i].getName() + "]");
                }

                info.setAction(newAction);
            }

            info.setTopic(getTopics());
        }
    }
}
