/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Author Luigi Zangrando <zangrando@pd.infn.it>
 *
 */

package org.glite.ce.monitor.action;

import org.glite.ce.monitorapij.queryprocessor.QueryResult;
import org.glite.ce.monitorapij.resource.types.Action;
import org.glite.ce.monitorapij.resource.types.Parameter;
import org.glite.ce.monitorapij.resource.types.Property;
import org.glite.ce.monitorapij.types.Event;
import org.glite.ce.monitorapij.types.Notification;

/**
 * This class is an extension of <code>Action</code> class.<br>
 * It is used to mark an ISM_LDIF formatted message as expired message. This is
 * done by adding a TTLCEinfo=0 field to the LDIF message, so as the purchaser
 * will interpret this message as expired.
 */
public class SendExpiredNotificationAction extends Action {
    private static final long serialVersionUID = 1L;



    /**
     * Creates a new SendExpiredNotificationAction object.
     */
    public SendExpiredNotificationAction() {
        super("SendExpiredNotification");

        Property property = new Property("supportedDialectsList", "LFID,ISM_LDIF,OLD_CLASSAD,NEW_CLASSAD,ISM_CLASSAD");
        setProperty(new Property[] {
            property
        });
    }



    /**
     * Check if this dialect is defined in to the supportedDialectsList
     * parameter.
     * 
     * @param dialect
     *            The dialect String to be checked.
     * @throws Exception
     *             This exception is thrown if the specified dialect is null or
     *             is not defined in to the supportedDialectsList parameter.
     */
    private void checkDialect(String dialect) throws Exception {
        if(dialect == null) {
            throw (new Exception("invalid dialect!"));
        }

        Property property = getProperty("supportedDialectsList");

        if(property != null) {
            String[] dialects = property.getValue().split(",");

            for (int i = 0; i < dialects.length; i++) {
                if(dialect.equalsIgnoreCase(dialects[i])) {
                    return;
                }
            }
        }
        throw (new Exception("invalid dialect! (\"" + dialect + "\")"));
    }



    /**
     * This method is called by <code>NotificationHolder</code> specifying a
     * <code>Notification</code> and a <code>QueryResult</code> array.<br>
     * It performs the specific operations of this action:<br>
     * The dialect is read from the topic of the specified notification, and
     * then it is checked using checkDialect() method to verify if it is a
     * ISM_LDIF dialect.<br>
     * Then, for each successful QueryResult, the relative sensor event is
     * called, and then its message is changed adding the TTLCEInfo information.
     * 
     * @param The
     *            specified notification.
     * @param result
     *            The specified <code>QueryResult</code> array.
     * @throws Exception
     *             DOCUMENT ME!
     */
    //public synchronized void execute(Notification notification, QueryResult[] result) throws Exception {

    public synchronized void execute() throws Exception {
        Parameter parameter = getParameter("notification");

        if(parameter == null) {
            return;
        }

        Notification notification = (Notification)parameter.getValue();

        parameter = getParameter("queryResult");

        if(parameter == null) {
            return;
        }
        
        QueryResult[] result = (QueryResult[])parameter.getValue();
        

        String dialect = notification.getTopic().getDialect(0).getName();

       // checkDialect(dialect);

        boolean isISM_LDIF = dialect.toLowerCase().indexOf("ldif") == -1 ? false : true;

        Event[] eventArray = notification.getEvent();
          
        for (int i = 0; i < eventArray.length; i++) {
            Event event = eventArray[i];
            QueryResult eventResult = result[i];
            
            for (int j = 0; j < eventResult.size(); j++) {
                if(eventResult.isSuccessfully(j) == this.isDoActionWhenQueryIs()) {
                    String msg = event.getMessage(j);

                    int index = msg.indexOf("TTLCEinfo");
                    if(index > -1) {
                        msg = msg.substring(0, index);
                    }
                    
                    if(isISM_LDIF) {
                        if(j > 0) {
                            msg += "\nTTLCEinfo: 0";
                        }
                    } else {
                        msg = msg.substring(0, msg.length() - 2);
                        msg += "\n\tTTLCEinfo = 0;\n]";
                    }
                    
                    event.setMessage(j, msg);
                }
            }
        }
    }
}
