/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Paolo Andreetto, <paolo.andreetto@pd.infn.it>
 *
 */

package org.glite.ce.monitor.configuration.basic;

import java.util.Hashtable;
import java.util.Iterator;

import javax.naming.Context;
import javax.naming.Name;
import javax.naming.NamingEnumeration;
import javax.naming.NameParser;
import javax.naming.NamingException;
import javax.naming.OperationNotSupportedException;
import javax.naming.directory.DirContext;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.BasicAttributes;
import javax.naming.directory.SearchControls;
import javax.naming.directory.ModificationItem;

import org.glite.ce.commonj.configuration.basic.ConfigItem;
import org.glite.ce.commonj.configuration.basic.ConfigNodeList;
import org.glite.ce.commonj.configuration.basic.ConfigDataBase;
import org.glite.ce.commonj.configuration.basic.ConfigNameParser;
import org.glite.ce.commonj.configuration.basic.ConfigQueryException;
import org.glite.ce.commonj.configuration.basic.NameClassPairEnumeration;
import org.glite.ce.commonj.configuration.basic.SearchResultEnumeration;

import org.apache.log4j.Logger;

public class CEMonConfigContext implements DirContext {

	private static Logger logger = Logger.getLogger(CEMonConfigContext.class.getName());

	protected final static NameParser nameParser = new ConfigNameParser();

	private ConfigDataBase configDB;

	public CEMonConfigContext(ConfigDataBase cdb) throws NamingException {
		if (cdb == null)
			throw new NamingException("CEMon configuration database undefined");
		configDB = cdb;

	}

	public Object clone() {
		try {
			return new CEMonConfigContext(configDB);
		} catch (NamingException nEx) {
			logger.error(nEx.getMessage());
		}

		return null;
	}

	public Attributes getAttributes(String name) throws NamingException {
		return getAttributes(name, null);
	}

	public Attributes getAttributes(Name name) throws NamingException {
		return getAttributes(name.toString(), null);
	}

	public Attributes getAttributes(String name, String[] ids) throws NamingException {
		Attributes answer = new BasicAttributes(false);

		try {
			ConfigItem node = configDB.query(name);
            
            if(node == null) {
                throw new ConfigQueryException("CEMonConfigContext:getAttributes() error: name \"" + name + "\" not found!");
            }

			if (ids == null) {
				Iterator items = node.keySet().iterator();
				while (items.hasNext()) {
					String tmps = (String) items.next();
					answer.put(tmps, node.get(tmps));
				}
			} else {
				for (int k = 0; k < ids.length; k++) {
					Object tmpo = node.get(ids[k]);
					if (tmpo != null)
						answer.put(ids[k], tmpo);
				}
			}
		} catch (ConfigQueryException cEx) {
			throw new NamingException(cEx.getMessage());
		}

		return answer;
	}

	public Attributes getAttributes(Name name, String[] ids) throws NamingException {
		return getAttributes(name.toString(), ids);
	}

	public Object lookup(Name name) throws NamingException {
		return lookup(name.toString());
	}

	public Object lookup(String name) throws NamingException {
		if (name.equals(""))
			return this.clone();

		ConfigItem node = null;
		try {
			node = configDB.query(name);
		} catch (ConfigQueryException cEx) {
			throw new NamingException(cEx.getMessage());
		}

		Object tmpo = node.getContent();

		if (tmpo instanceof ConfigNodeList) {
			throw new NamingException("Subcontext not yet available");
		}

		return tmpo;
	}

	public void bind(Name name, Object obj) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void bind(String name, Object obj) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void rebind(Name name, Object obj) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void rebind(String name, Object obj) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void unbind(Name name) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void unbind(String name) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void rename(Name oldName, Name newName) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void rename(String oldName, String newName) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public NamingEnumeration list(Name name) throws NamingException {
		return list(name.toString());
	}

	public NamingEnumeration list(String name) throws NamingException {
		ConfigItem node = null;
		try {
			node = configDB.query(name);
		} catch (ConfigQueryException cEx) {
			throw new NamingException(cEx.getMessage());
		}

		return new NameClassPairEnumeration(node);
	}

	public NamingEnumeration listBindings(Name name) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public NamingEnumeration listBindings(String name) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void destroySubcontext(Name name) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void destroySubcontext(String name) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public Context createSubcontext(Name name) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public Context createSubcontext(String name) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public Object lookupLink(Name name) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public Object lookupLink(String name) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public NameParser getNameParser(Name name) throws NamingException {
		return nameParser;
	}

	public NameParser getNameParser(String name) throws NamingException {
		return nameParser;
	}

	public String composeName(String name, String prefix)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public Name composeName(Name name, Name prefix) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public Object addToEnvironment(String propName, Object propVal)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public Object removeFromEnvironment(String propName) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public Hashtable getEnvironment() throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void close() throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void modifyAttributes(Name name, int modOp, Attributes attrs)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void modifyAttributes(String name, int modOp, Attributes attrs)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void modifyAttributes(Name name, ModificationItem[] mods)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void modifyAttributes(String name, ModificationItem[] mods)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void bind(Name name, Object obj, Attributes attrs)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void bind(String name, Object obj, Attributes attrs)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void rebind(Name name, Object obj, Attributes attrs)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public void rebind(String name, Object obj, Attributes attrs)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public DirContext createSubcontext(Name name, Attributes attrs)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public DirContext createSubcontext(String name, Attributes attrs)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public DirContext getSchema(Name name) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public DirContext getSchema(String name) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public DirContext getSchemaClassDefinition(Name name)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public DirContext getSchemaClassDefinition(String name)
			throws NamingException {
		throw new OperationNotSupportedException();
	}

	public NamingEnumeration search(Name name, Attributes matchingAttributes,
			String[] attributesToReturn) throws NamingException {
		return search(name.toString(), matchingAttributes, attributesToReturn);
	}

	public NamingEnumeration search(String name, Attributes matchingAttributes,
			String[] attributesToReturn) throws NamingException {

		String[] attrNames = new String[matchingAttributes.size()];
		Object[] attrValues = new Object[matchingAttributes.size()];

		NamingEnumeration allAttrs = matchingAttributes.getAll();
		for (int k = 0; allAttrs.hasMore(); k++) {
			Attribute attr = (Attribute) allAttrs.next();
			attrNames[k] = attr.getID();
			attrValues[k] = attr.get(0);
		}

		ConfigItem[] nodes = null;
		try {
			nodes = configDB.query(name, attrNames, attrValues);
		} catch (ConfigQueryException cEx) {
			throw new NamingException(cEx.getMessage());
		}

		return new SearchResultEnumeration(nodes);

	}

	public NamingEnumeration search(Name name, Attributes matchingAttributes)
			throws NamingException {
		return search(name.toString(), matchingAttributes, null);
	}

	public NamingEnumeration search(String name, Attributes matchingAttributes)
			throws NamingException {
		return search(name, matchingAttributes, null);
	}

	public NamingEnumeration search(Name name, String filter,
			SearchControls cons) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public NamingEnumeration search(String name, String filter,
			SearchControls cons) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public NamingEnumeration search(Name name, String filterExpr,
			Object[] filterArgs, SearchControls cons) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public NamingEnumeration search(String name, String filterExpr,
			Object[] filterArgs, SearchControls cons) throws NamingException {
		throw new OperationNotSupportedException();
	}

	public String getNameInNamespace() throws NamingException {
		throw new OperationNotSupportedException();
	}

}
