/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Paolo Andreetto, <paolo.andreetto@pd.infn.it>
 *
 */

package org.glite.ce.monitor.configuration.basic;

import java.util.HashMap;
import java.util.Iterator;
import java.io.File;
import org.xml.sax.SAXParseException;
import org.xml.sax.Attributes;

import org.glite.ce.commonj.configuration.basic.ConfigItem;
import org.glite.ce.commonj.configuration.basic.ConfigNodes;
import org.glite.ce.commonj.configuration.basic.ConfigParser;
import org.glite.ce.commonj.configuration.basic.TriggerManager;

import org.glite.ce.monitorapij.resource.types.CEMonResource;
import org.glite.ce.monitorapij.resource.types.Property;

import java.net.URI;
import java.net.URISyntaxException;

import org.apache.axis.types.URI.MalformedURIException;
import org.apache.log4j.Logger;

public class CEMonConfigParser implements ConfigParser {

    private final static Logger logger = Logger.getLogger(CEMonConfigParser.class.getName());

    private static String[] SERVICE_ATTRIBUTES = { "sslkeyfile", "sslkeypasswd", "sslcertfile",
						   "gridproxyfile", "sslCAfiles", "sslCRLfiles", 
						   "ServiceDescription", "ServiceVersion", "InterfaceVersion", "notificationSharedThread",
						   "notificationTimeout_sec", "maxSubscriptionExpirationTime_hours",
						   "backendLocation", "backendCacheSize",
						   "confProvider", "subscriptionRegistryConfProvider" };

    private HashMap currentParams;

    private TriggerManager triggerManager;

    public CEMonConfigParser(TriggerManager tMan) {
        triggerManager = tMan;
    }

    public String getObjectName(String uri, String name, String qName, Attributes attributes)
            throws SAXParseException {
        if (qName.equals("service") || qName.equals("action") || qName.equals("queryprocessor")
                || qName.equals("sensor")) {
            String result = attributes.getValue("id");
            if (result == null)
                throw new SAXParseException("Missing service name", null);

            return result;
        }

        return null;
    }

    public void startElement(String uri, String name, String qName, Attributes attributes,
            ConfigItem prevObj) throws SAXParseException {

        if (qName.equals("service")) {

            Long tmpl = new Long(System.currentTimeMillis());
            boolean configChanged = false;

            ConfigNodes currentList = (ConfigNodes) prevObj.getContent();
            if (currentList == null) {

                currentList = new ConfigNodes();
                prevObj.setContent(currentList);
                prevObj.put("creationtime", tmpl);
                prevObj.put("category", "service");

            }

            for (int k = 0; k < SERVICE_ATTRIBUTES.length; k++) {

                String tmp1 = attributes.getValue(SERVICE_ATTRIBUTES[k]);
                String tmp2 = (String) prevObj.get(SERVICE_ATTRIBUTES[k]);

                if (tmp1 == null) {
                    if (tmp2 != null) {
                        prevObj.remove(SERVICE_ATTRIBUTES[k]);
                        configChanged = true;
                    }
                } else if (tmp2 == null || !tmp1.equalsIgnoreCase(tmp2)) {
                    prevObj.put(SERVICE_ATTRIBUTES[k], tmp1);
                    configChanged = true;
                }
            }

            if (configChanged) {
                prevObj.put("modificationtime", tmpl);
            }
        } else if (qName.equals("action") || qName.equals("queryprocessor")
                || qName.equals("sensor")) {
            Long tmpl = new Long(System.currentTimeMillis());
            boolean configChanged = false;

            currentParams = new HashMap();

            CEMonResource resource = (CEMonResource) prevObj.getContent();
            if (resource == null) {
                resource = new CEMonResource();
                resource.setId(attributes.getValue("id"));
                prevObj.setContent(resource);
                prevObj.put("creationtime", tmpl);
                prevObj.put("category", qName);
            }

            String tmps = attributes.getValue("name");
            if (tmps == null)
                throw new SAXParseException("Missing resource name", null);
            if (!tmps.equals(resource.getName())) {
                resource.setName(tmps);
                configChanged = true;
            }

            tmps = attributes.getValue("type");
            if (tmps == null)
                throw new SAXParseException("Missing type", null);
            if (!tmps.equals(resource.getType())) {
                resource.setType(tmps);
                configChanged = true;
            }

            tmps = attributes.getValue("jarpath");
            if (tmps == null)
                throw new SAXParseException("Missing jarpath", null);

            try {
                URI newURI = new URI(tmps);
                URI oldURI = resource.getJarPath();
                File newJarFile = tmps.startsWith("file:") ? new File(tmps.substring(5)) : new File(tmps);

                if (!newURI.equals(oldURI)) {
                    resource.setJarPath(newURI);
                    configChanged = true;
                    if (oldURI != null)
                        triggerManager.removeTrigger(new File(oldURI.toString()));
                    triggerManager.addTrigger(newJarFile);
                } else {
                    long jarLastModTime = newJarFile.lastModified();
                    Object tmpo = prevObj.get("modificationtime");
                    if (tmpo == null || ((Long) tmpo).longValue() < jarLastModTime) {
                        configChanged = true;
                    }
                }

            } catch (URISyntaxException e) {
                throw new SAXParseException(e.getMessage(), null);
            }

            if (configChanged) {
                prevObj.put("modificationtime", tmpl);
            }

        } else if (qName.equals("property")) {

            String pName = attributes.getValue("name");
            String pValue = attributes.getValue("value");
            if (pName == null || pValue == null)
                throw new SAXParseException("Bad parameter definition", null);

            currentParams.put(pName.trim(), pValue.trim());
        }
    }

    public void endElement(String uri, String name, String qName, ConfigItem currentObj) {

        if (qName.equals("action") || qName.equals("queryprocessor") || qName.equals("sensor")) {

            Long tmpl = new Long(System.currentTimeMillis());
            boolean configUnchanged = true;

            CEMonResource resource = (CEMonResource) currentObj.getContent();
            Property[] props = resource.getProperty();
            if (props != null && props.length == currentParams.size()) {

                for (int k = 0; k < props.length && configUnchanged; k++) {

                    String tmps = (String) currentParams.get(props[k].getName());
                    configUnchanged = configUnchanged && tmps != null
                            && tmps.equalsIgnoreCase(props[k].getValue());
                }

            } else {
                configUnchanged = false;
            }

            if (!configUnchanged) {
                Property[] newProps = new Property[currentParams.size()];

                Iterator items = currentParams.keySet().iterator();
                for (int k = 0; items.hasNext(); k++) {
                    String tmpName = (String) items.next();
                    newProps[k] = new Property(tmpName, (String) currentParams.get(tmpName));
                }

                resource.setProperty(newProps);
                currentObj.put("modificationtime", tmpl);
            }
        }
    }

    public void characters(char[] chars, int start, int length, ConfigItem currentContext)
            throws SAXParseException {}

}
