/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Paolo Andreetto, <paolo.andreetto@pd.infn.it>
 *
 */

package org.glite.ce.monitor.configuration.basic;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Calendar;
import org.xml.sax.SAXParseException;
import org.xml.sax.Attributes;

import org.glite.ce.commonj.configuration.basic.ConfigItem;
import org.glite.ce.commonj.configuration.basic.ConfigParser;
import org.glite.ce.commonj.configuration.basic.TriggerManager;

import org.glite.ce.monitorapij.resource.types.Topic;
import org.glite.ce.monitorapij.resource.types.Policy;
import org.glite.ce.monitorapij.resource.types.Dialect;
import org.glite.ce.monitorapij.resource.types.Query;
import org.glite.ce.monitorapij.resource.types.Action;
import org.glite.ce.monitorapij.resource.types.SubscriptionPersistent;


import org.apache.log4j.Logger;

public class SubscriptionConfigParser implements ConfigParser {
    private static Logger logger = Logger.getLogger(SubscriptionConfigParser.class.getName());

    private HashMap dialects;

    private HashSet actions;

    private String token;

    private boolean isChanged;

    public SubscriptionConfigParser(TriggerManager tMan) {}

    public String getObjectName(String uri, String name, String qName, Attributes attributes)
            throws SAXParseException {
        if (qName.equals("subscription")) {
            String result = attributes.getValue("id");
            if (result == null)
                throw new SAXParseException("Missing subscription name", null);

            return result;
        }

        return null;
    }

    public void startElement(String uri, String name, String qName, Attributes attributes,
            ConfigItem prevObj) throws SAXParseException {

        SubscriptionPersistent currSubscr = (SubscriptionPersistent) prevObj.getContent();
        
        if (currSubscr == null) {
            currSubscr = new SubscriptionPersistent();

            Calendar newCalendar = Calendar.getInstance();
            newCalendar.setTimeInMillis(4398046511104L);
            currSubscr.setExpirationTime(newCalendar);

            prevObj.setContent(currSubscr);
            prevObj.put("creationtime", new Long(System.currentTimeMillis()));
            prevObj.put("category", "subscription");
        }

        if (qName.equals("subscription")) {
            isChanged = false;

            try {
                String tmps = attributes.getValue("id");
                if (tmps != null && !currSubscr.equals(tmps)) {
                    currSubscr.setId(tmps);
                    currSubscr.setName(tmps);
                    isChanged = true;
                    logger.debug("Changed id");
                }

                tmps = attributes.getValue("monitorConsumerURL");
                               
                if (tmps == null)
                    throw new SAXParseException("Missing consumer URL", null);
                
                URI newURL = new URI(tmps);                
                URI oldURL = currSubscr.getMonitorConsumerURL();
                
                if (oldURL == null || !oldURL.equals(newURL)) {
                    currSubscr.setMonitorConsumerURL(newURL);
                    isChanged = true;
                    logger.debug("Changed monitorConsumerURL");
                }

                tmps = attributes.getValue("retryCount");
                int currRetryCount = (tmps != null) ? Integer.parseInt(tmps) : -1;
                int oldCount = currSubscr.getMaxRetryCount();
                if (oldCount != currRetryCount) {
                    currSubscr.setMaxRetryCount(currRetryCount);
                    isChanged = true;
                    logger.debug("Changed retryCount");
                }

                tmps = attributes.getValue("credentialfile");
                if (tmps != null && !tmps.equals(currSubscr.getCredentialFile())) {
                    currSubscr.setCredentialFile(tmps);
                    isChanged = true;
                    logger.debug("Changed credentialfile");
                }

                tmps = attributes.getValue("sslkeypasswd");
                if (tmps != null && !tmps.equals(currSubscr.getPassphrase())) {
                    currSubscr.setPassphrase(tmps);
                    isChanged = true;
                    logger.debug("Changed sslkeypasswd");
                }

                tmps = attributes.getValue("sslprotocol");
                if (tmps != null && !tmps.equals(currSubscr.getSSLProtocol())) {
                    currSubscr.setSSLProtocol(tmps);
                    isChanged = true;
                    logger.debug("Changed sslprotocol");
                }
                
                tmps = attributes.getValue("subscriberId");
                if (tmps != null && !tmps.equals(currSubscr.getSubscriberId())) {
                    currSubscr.setSubscriberId(tmps);
                    isChanged = true;
                    logger.debug("Changed subscriberId");
                }
                
                tmps = attributes.getValue("subscriberGroup");
                if (tmps != null && !tmps.equals(currSubscr.getSubscriberGroup())) {
                    currSubscr.setSubscriberGroup(tmps);
                    isChanged = true;
                    logger.debug("Changed subscriberGroup");
                }

            } catch (NumberFormatException numEx) {
                throw new SAXParseException(numEx.getMessage(), null);
            } catch (URISyntaxException e) {
                throw new SAXParseException(e.getMessage(), null);
            }

        } else if (qName.equals("topic")) {
            Topic currTopic = currSubscr.getTopic();
            if (currTopic == null) {
                currTopic = new Topic();
                currSubscr.setTopic(currTopic);
            }

            String tmps = attributes.getValue("name");
            if (tmps == null)
                throw new SAXParseException("Missing topic name", null);
            if (!tmps.equals(currTopic.getName())) {
                currTopic.setName(tmps);
                isChanged = true;
                logger.debug("Changed topic");
            }

            dialects = new HashMap();

        } else if (qName.equals("dialect")) {
            String tmps = attributes.getValue("name");
            if (tmps == null)
                throw new SAXParseException("Missing dialect name", null);
            dialects.put(tmps, new Dialect(tmps));

        } else if (qName.equals("policy")) {

            Policy currPolicy = currSubscr.getPolicy();
            if (currPolicy == null) {
                currPolicy = new Policy();
                currSubscr.setPolicy(currPolicy);
            }

            String tmps = attributes.getValue("rate");
            if (tmps == null)
                throw new SAXParseException("Missing policy rate", null);
            try {
                int rate = Integer.parseInt(tmps);
                if (rate != currPolicy.getRate()) {
                    currPolicy.setRate(rate);
                    isChanged = true;
                    logger.debug("Changed rate");
                }
            } catch (NumberFormatException numEx) {
                throw new SAXParseException(numEx.getMessage(), null);
            }

            actions = new HashSet();

        } else if (qName.equals("query")) {
            Policy currPolicy = currSubscr.getPolicy();
            Query currQuery = currPolicy.getQuery();
            if (currQuery == null) {
                currQuery = new Query();
                currPolicy.setQuery(currQuery);
            }

            String tmps = attributes.getValue("queryLanguage");
            if (tmps == null)
                throw new SAXParseException("Missing query language", null);
            if (!tmps.equals(currQuery.getQueryLanguage())) {
                currQuery.setQueryLanguage(tmps);
                isChanged = true;
                logger.debug("Changed queryLanguage");
            }

        } else if (qName.equals("action")) {
            String tmps = attributes.getValue("name");
            if (tmps == null)
                throw new SAXParseException("Missing action name", null);
            ActionReference action = new ActionReference(tmps);

            tmps = attributes.getValue("doActionWhenQueryIs");
            if (tmps == null)
                throw new SAXParseException("Missing action condition", null);
            action.setDoActionWhenQueryIs(tmps.equalsIgnoreCase("true"));

            actions.add(action);
        }

    }

    public void endElement(String uri, String name, String qName, ConfigItem currentObj)
            throws SAXParseException {

        SubscriptionPersistent currSubscr = (SubscriptionPersistent) currentObj.getContent();

        if (qName.equals("subscription")) {
            if (isChanged) {
                currentObj.put("modificationtime", new Long(System.currentTimeMillis()));
            }
        } else 
        if (qName.equals("topic")) {
            Topic currTopic = currSubscr.getTopic();
            Dialect[] allDialects = currTopic.getDialect();
            boolean mustChange = false;

            if (allDialects != null && allDialects.length == dialects.size()) {
                for (int k = 0; k < allDialects.length && !mustChange; k++) {
                    /*
                     * We don't check if queryLanguages are equals
                     */
                    Dialect dialect = (Dialect) dialects.get(allDialects[k].getName());
                    mustChange = mustChange | (dialect == null);
                }
            } else {
                mustChange = true;
            }

            if (mustChange) {
                allDialects = new Dialect[dialects.size()];
                dialects.values().toArray(allDialects);
                currTopic.setDialect(allDialects);
                isChanged = true;
                logger.debug("Changed dialects");
            }

        } else if (qName.equals("policy")) {
            Policy currPolicy = currSubscr.getPolicy();
            Action[] allActions = currPolicy.getAction();
            boolean mustChange = false;

            if (allActions != null && allActions.length == actions.size()) {
                for (int k = 0; k < allActions.length && !mustChange; k++) {
                    mustChange = mustChange | !actions.contains(allActions[k]);
                    logger.debug("Testing " + allActions[k].getName() + " " + mustChange);
                }
            } else {
                mustChange = true;
                logger.debug("Mustchange true");
            }

            if (mustChange) {
                allActions = new Action[actions.size()];
                actions.toArray(allActions);
                currPolicy.setAction(allActions);
                isChanged = true;
                logger.debug("Changed actions");
            }
        } else if (qName.equals("query")) {
            Policy currPolicy = currSubscr.getPolicy();
            Query currQuery = currPolicy.getQuery();
            if (token == null || token.length() == 0)
                throw new SAXParseException("Missing query expression", null);

            if (!token.equals(currQuery.getExpression())) {
                currQuery.setExpression(token);
                isChanged = true;
                logger.debug("Changed expression");
            }
        }

        token = null;
    }

    public void characters(char[] chars, int start, int length, ConfigItem currentContext)
            throws SAXParseException {
        token = (new String(chars, start, length)).trim();
    }

    public class ActionReference extends Action {

        public ActionReference(String name) {
            super();
            this.setName(name);
        }

        public boolean equals(Object obj) {
            if (obj == null || !(obj instanceof Action))
                return false;

            Action action = (Action) obj;
            return this.getName().equals(action.getName())
                    && this.isDoActionWhenQueryIs() == action.isDoActionWhenQueryIs();

        }
    }
}
