/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Luigi Zangrando <zangrando@pd.infn.it>
 *
 */

package org.glite.ce.monitor.holder;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Properties;

import javax.naming.Context;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.xml.rpc.ServiceException;

import org.apache.axis.AxisFault;
import org.apache.axis.Handler;
import org.apache.axis.SimpleChain;
import org.apache.axis.SimpleTargetedChain;
import org.apache.axis.configuration.SimpleProvider;
import org.apache.axis.transport.http.HTTPTransport;
import org.apache.axis.transport.http.SocketHolder;
import org.apache.log4j.Logger;
import org.glite.ce.commonj.utils.BooleanLock;
import org.glite.ce.commonj.utils.CEUtils;
import org.glite.ce.monitorapij.queryprocessor.QueryProcessor;
import org.glite.ce.monitorapij.queryprocessor.QueryResult;
import org.glite.ce.monitorapij.resource.Resource;
import org.glite.ce.monitorapij.resource.ResourceEvent;
import org.glite.ce.monitorapij.resource.ResourceHolder;
import org.glite.ce.monitorapij.resource.types.Action;
import org.glite.ce.monitorapij.resource.types.Dialect;
import org.glite.ce.monitorapij.resource.types.Parameter;
import org.glite.ce.monitorapij.resource.types.Policy;
import org.glite.ce.monitorapij.resource.types.Query;
import org.glite.ce.monitorapij.resource.types.SubscriptionPersistent;
import org.glite.ce.monitorapij.resource.types.Topic;
import org.glite.ce.monitorapij.sensor.SensorEvent;
import org.glite.ce.monitorapij.sensor.SensorOutputDataFormat;
import org.glite.ce.monitorapij.types.Event;
import org.glite.ce.monitorapij.types.Notification;
import org.glite.ce.monitorapij.ws.CEMonitorConsumerLocator;
import org.glite.ce.monitorapij.ws.CEMonitorConsumerPort;
import org.glite.security.trustmanager.ContextWrapper;

/**
 * This classed is used to take care of all aspects relating notifications. It
 * uses a set of classes to process a query request for sensor events, apply
 * predefined dialects and actions to these events, and hold a list of listeners
 * subscribed for notifications. Finally it submits the opportunely formatted
 * notifications.
 */
public final class NotificationHolder extends ResourceHolder {
    private final static Logger logger = Logger.getLogger(NotificationHolder.class.getName());
    private TopicHolder topicHolder;
    private int rate = 0;
    private QueryProcessorHolder queryProcessorHolder;
    private ActionHolder actionHolder;
    private final Object lock = new Object();
    private boolean exit = false;
    private BooleanLock bl = null;

    // private static DirContext rootContext = null;
    private static Attributes serviceAttrs = null;

    /**
     * Creates a new NotificationHolder object specifying the rate and the
     * <code>TopicHolder</code>.
     * 
     * @param rate
     *            The subscription rate.
     * @param topicHolder
     *            The <code>TopicHolder</code>.
     */
    public NotificationHolder(int rate, TopicHolder topicHolder) {
        this("" + rate, rate, topicHolder);
    }

    public NotificationHolder(String name, int rate, TopicHolder topicHolder) throws IllegalArgumentException {
        super();

        if (name == null) {
            throw new IllegalArgumentException("name not specified!");
        }
        
        setName(name);
        
        if (rate <= 0) {
            throw new IllegalArgumentException("rate must be > 0!");
        }

        if (topicHolder == null) {
            throw new IllegalArgumentException("topicHolder not specified!");
        }

        this.rate = rate;
        this.topicHolder = topicHolder;
        this.bl = new BooleanLock(true);

        if (serviceAttrs == null) {
            String confProvider = CEUtils.getConfigurationProvider(CEUtils.CEMONITOR_LABEL);
            String confProviderURL = CEUtils.getConfigurationURL(CEUtils.CEMONITOR_LABEL);

            if (confProvider != null && confProviderURL != null) {
                Hashtable<String, String> jndiProperties = new Hashtable<String, String>(2);
                jndiProperties.put(Context.INITIAL_CONTEXT_FACTORY, confProvider);
                jndiProperties.put(Context.PROVIDER_URL, confProviderURL);

                try {
                    DirContext rootContext = (DirContext) new InitialDirContext(jndiProperties);
                    serviceAttrs = rootContext.getAttributes("");
                    rootContext = null;
                } catch (NamingException nEx) {
                    logger.error(nEx.getMessage(), nEx);
                }
            }
        }
        
        Thread thread = new Thread(this);
        thread.start();
        thread.setName(name);
    }

    public String getCategory() {
        return "NotificationHolder";
    }

    /**
     * Set the <code>ActionHolder</code>.
     * 
     * @param ah
     *            The <code>ActionHolder</code>.
     */
    public void setActionHolder(ActionHolder ah) {
        actionHolder = ah;
    }

    /**
     * Set the <code>QueryProcessorHolder</code>.
     * 
     * @param qph
     *            The <code>QueryProcessorHolder</code>.
     */
    public void setQueryProcessorHolder(QueryProcessorHolder qph) {
        queryProcessorHolder = qph;
    }

    /**
     * Get the number of registered subscriptions.
     * 
     * @return The size of the subscription map.
     */
    public int getSubscriptionListSize() {
        return getResourceListSize();
        
    }

    /**
     * Add a <code>SubscriptionPersistent</code> to the list of the registered
     * subscriptions.
     * 
     * @param subscription
     *            The <code>SubscriptionPersistent</code> to add.
     */
    public void addSubscription(SubscriptionPersistent subscription) {
        if (subscription != null) {
            synchronized (lock) {
                addResource(subscription);
                lock.notifyAll();
                bl.setValue(false);
            }
        }
    }

    /**
     * Remove a <code>SubscriptionPersistent</code> from the list of the
     * registered subscriptions.
     * 
     * @param subscription
     *            The <code>SubscriptionPersistent</code> to remove.
     */
    public void removeSubscription(SubscriptionPersistent subscription) {
        if (subscription != null) {
            synchronized (lock) {
                removeResource(subscription);
                lock.notifyAll();

                if (getResourceListSize() == 0) {
                    bl.setValue(true);
                }
            }
        }
    }

    private void processPolicy(SensorEventArrayList eventList, Policy policy, Notification notification, String subscriptionId) throws Exception {
        if (eventList == null || policy == null || notification == null) {
            return;
        }

        Query query = policy.getQuery();
        QueryProcessor qp = null;

        Action[] actions = policy.getAction();
        if (actions == null) {
            // logger.error("processPolicy() - policy.actions not found!")
            throw (new Exception("policy.actions not found!"));
        }

        if (query != null) {
            if (queryProcessorHolder == null) {
                throw (new Exception("queryProcessorHolder is null!"));
            }

            qp = queryProcessorHolder.getQueryProcessor(query.getQueryLanguage());

            if (qp == null) {
                logger.error("QueryProcessor \"" + query.getQueryLanguage() + "\" not found!");
                throw (new Exception("QueryProcessor \"" + query.getQueryLanguage() + "\" not found!"));
            }
        }
        org.glite.ce.monitorapij.types.Dialect[] dialectArray = notification.getTopic().getDialect();
        String dataFormat = "default";

        if ((dialectArray != null) && dialectArray.length > 0 && dialectArray[0].getName() != null) {
            dataFormat = dialectArray[0].getName();
        }

        ArrayList<Event> notificationEventList = new ArrayList<Event>(0);
        ArrayList<QueryResult> queryResultList = new ArrayList<QueryResult>(0);

        for (int i = 0; i < eventList.size(); i++) {
            SensorEvent event = (SensorEvent) eventList.get(i);

            if (event != null)
                if (event.isExpired()) {
                    try {
                        logger.debug("removing expired event [id=" + event.getID() + "] [name=" + event.getName() + "] [expirationTime=" + event.getExpirationTime() + "]");
                        topicHolder.removeEvent(event);
                    } catch (Throwable t) {
                        logger.error("cannot remove the event [id=" + event.getID() + "] [name=" + event.getName() + "]!", t);
                    }

                    eventList.remove(i);
                } else {
                    SensorOutputDataFormat format = event.getSensorOutputDataFormatApplied();

                    if (format == null || !format.getName().equals(dataFormat)) {
                        event.applyFormat(dataFormat);
                    }

                    if (qp != null) {
                        queryResultList.add(qp.evaluate(query, event));
                    } else {
                        QueryResult queryRes = new QueryResult(event.getMessage().length);
                        for (int x = 0; x < queryRes.size(); x++) {
                            queryRes.setResult(x, true);
                        }

                        queryResultList.add(queryRes);
                    }

                    if (event.getMessage() != null && event.getMessage().length > 0) {
                        notificationEventList.add(new Event(event.getID(), event.getTimestamp(), event.getMessage(), event.getProducer()));
                    }
                }
        }

        notificationEventList.trimToSize();
        Event[] eventArray = new Event[notificationEventList.size()];
        eventArray = (Event[]) notificationEventList.toArray(eventArray);
        notificationEventList.clear();

        notification.setEvent(eventArray);

        QueryResult[] queryResult = new QueryResult[notificationEventList.size()];
        queryResult = (QueryResult[]) queryResultList.toArray(queryResult);

        for (int i = 0; i < actions.length; i++) {
            try {
                Action action = actionHolder.getAction(actions[i].getName());

                if (action == null) {
                    logger.debug("Action \"" + actions[i].getName() + "\" not found!");

                    continue;
                }

                action.setDoActionWhenQueryIs(actions[i].isDoActionWhenQueryIs());
                action.addParameter(new Parameter("notification", notification));
                action.addParameter(new Parameter("queryResult", queryResult));
                action.addParameter(new Parameter("subscriptionId", subscriptionId));
                action.execute();
            } catch (Exception e) {
                logger.error("processPolicy() - Exception catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
            }
        }
    }

    /**
     * This method is used by a scheduled execution of this class.The operations
     * performed are:<br>
     * 1) For each registered subscription a control is made to verify if this
     * is expired. If this is the case the subscription is unregistered.<br>
     * 2) If the subscription is not expired or paused then the query is
     * processed on the event, then actions on notification are performed based
     * on the query processing results, and finally the consumer or purchaser
     * registered for that notification is notified. <br>
     */
    public void run() {
        while (!exit) {
            CEMonitorConsumerPort consumer = null;
            SSLConfigSender sslConfigSender = null;

            List<Resource> subscriptionList = null;
            ArrayList<Resource> updatedSubscriptionList = new ArrayList<Resource>(0);
            ArrayList<Resource> unregisteredSubscriptionList = new ArrayList<Resource>(0);
            SubscriptionPersistent subscription = null;

            synchronized (lock) {
                subscriptionList = getResourceList();
            }

            for (Resource resource : subscriptionList) {
                if (exit) {
                    break;
                }
                
                subscription = (SubscriptionPersistent)resource;

                if (subscription.isExpired()) {
                    logger.info("[name=" + getName() + "] - unregistering subscription " + subscription.getId() + " [reason: the subscription is expired!]");
                    unregisteredSubscriptionList.add(subscription);
                } else if (!subscription.isPaused()) {
                    URI consumerURI = subscription.getMonitorConsumerURL();

                    if (consumerURI == null) {
                        logger.info("[name=" + getName() + "] - unregistering subscription " + subscription.getId() + " [reason: consumer URL not found!]");
                        unregisteredSubscriptionList.add(subscription);
                    } else {
                        try {
                            logger.debug("[name=" + getName() + "] - processing subscription id " + subscription.getId());

                            Topic topic = subscription.getTopic();
                            org.glite.ce.monitorapij.types.Topic newTopic = new org.glite.ce.monitorapij.types.Topic();
                            newTopic.setName(topic.getName());

                            if (topic.getDialect() != null && topic.getDialect().length > 0) {
                                Dialect[] dialect = topic.getDialect();
                                org.glite.ce.monitorapij.types.Dialect[] newDialect = new org.glite.ce.monitorapij.types.Dialect[dialect.length];

                                for (int x = 0; x < dialect.length; x++) {
                                    newDialect[x] = new org.glite.ce.monitorapij.types.Dialect(dialect[x].getName(), dialect[x].getQueryLanguage());
                                }
                                newTopic.setDialect(newDialect);
                            }

                            Notification notification = new Notification(subscription.getExpirationTime(), newTopic, null, new org.apache.axis.types.URI(consumerURI.toString()));

                            Policy policy = subscription.getPolicy();
                            SensorEventArrayList event = topicHolder.getEvents(topic.getName(), subscription.getSubscriberId(), subscription.getSubscriberGroup());

                            logger.debug("[name=" + getName() + "] - found " + event.size() + " events");

                            boolean done = false;

                            while (!done) {
                                SensorEventArrayList subEventList = event.drain(100);

                                if (policy != null) {
                                    processPolicy(subEventList, policy, notification, subscription.getId());
                                }

                                if (notification.getEvent().length > 0) {
                                    logger.info("[name=" + getName() + "] - sending notification (containing " + notification.getEvent().length + " events) to " + new URL(consumerURI.toString()) + " ...");

                                    if (consumer == null) {
                                        sslConfigSender = getSSLConfigSender(subscription.getCredentialFile(), subscription.getPassphrase(), subscription.getSSLProtocol());

                                        consumer = getCEMonitorConsumerPort(consumerURI.toString(), sslConfigSender);
                                    }

                                    if (logger.isDebugEnabled()) {
                                        printNotification(notification, subEventList, subscription.getId(), subscription.getSubscriberId());
                                    }

                                    consumer.notify(notification);
                                    ((org.apache.axis.client.Stub) consumer).clearAttachments();
                                    ((org.apache.axis.client.Stub) consumer).clearHeaders();
                                    ((org.apache.axis.client.Stub) consumer)._getCall().clearOperation();
                                    ((org.apache.axis.client.Stub) consumer)._getCall().clearHeaders();
                                    ((org.apache.axis.client.Stub) consumer)._getCall().getMessageContext().dispose();
                                    ((org.apache.axis.client.Stub) consumer)._getCall().getMessageContext().getAxisEngine().cleanup();

                                    logger.info("[name=" + getName() + "] - [done]");
                                } else {
                                    logger.info("[name=" + getName() + "] - the notification doesn't contains messages to be notified! [aborted]");
                                }

                                if (subscription.getMaxRetryCount() != -1) {
                                    subscription.resetRetryCount();

                                    try {
                                        updatedSubscriptionList.add(subscription);
                                    } catch (Exception e1) {
                                        logger.error(e1);
                                    }
                                }

                                subEventList = null;
                                if (event.size() == 0) {
                                    done = true;
                                }
                            }

                            notification = null;
                            consumer = null;
                        } catch (MalformedURLException e) {
                            logger.info("[name=" + getName() + "] - unregistering subscription " + subscription.getId() + " [reason: consumer URL malformed (" + subscription.getMonitorConsumerURL() + ")]");
                            unregisteredSubscriptionList.add(subscription);
                        } catch (ServiceException e) {
                            logger.error("[name=" + getName() + "] - ServiceException occurred: subscription id = " + subscription.getId() + " consumer URL = " + subscription.getMonitorConsumerURL()
                                    + " message error = " + e.getMessage());
                        } catch (NamingException e) {
                            logger.error("[name=" + getName() + "] - NamingException occurred: subscription id = " + subscription.getId() + " consumer URL = " + subscription.getMonitorConsumerURL() + " message error = "
                                    + e.getMessage());
                        } catch (IOException e) {
                            logger.error("[name=" + getName() + "] - IOException occurred: subscription id = " + subscription.getId() + " consumer URL = " + subscription.getMonitorConsumerURL() + " message error = "
                                    + e.getMessage());

                            if (subscription.getMaxRetryCount() != -1) {
                                int retries = subscription.decrRetryCount();

                                logger.info("[name=" + getName() + "] - decrementing the retry count to " + retries + " for the subscription id = " + subscription.getId() + "; consumer URL = "
                                        + subscription.getMonitorConsumerURL() + "; message error = " + e.getMessage());

                                if (retries == 0) {
                                    logger.info("[name=" + getName() + "] - unregistering subscription " + subscription.getId() + " [reason: retry count = 0]");
                                    unregisteredSubscriptionList.add(subscription);
                                } else {
                                    try {
                                        updatedSubscriptionList.add(subscription);
                                    } catch (Exception e1) {
                                        logger.error("[name=" + getName() + "] - Exception catched: [" + e1.getMessage() + "]. Follow the Stack trace:", e1);
                                    }
                                }
                            }
                        } catch (Exception e) {
                            logger.error("[name=" + getName() + "] - Exception catched: subscriptionId = " + subscription.getId() + " consumer URL = " + subscription.getMonitorConsumerURL() + " message error = "
                                    + e.getMessage() + ". Follow the Stack trace:", e);
                        } finally {
                            if (sslConfigSender != null) {
                                SocketHolder sHolder = sslConfigSender.getSocketHolder();
                                if (sHolder != null && sHolder.getSocket() != null) {
                                    try {
                                        sHolder.getSocket().close();
                                    } catch (Exception e) {
                                        logger.error("[name=" + getName() + "] - Exception catched: [" + e.getMessage() + "]. Follow the Stack trace:", e);
                                    }
                                }
                            }
                        }
                    }
                }

                fireEvent(ResourceEvent.RESOURCE_REMOVED, unregisteredSubscriptionList);
                fireEvent(ResourceEvent.RESOURCE_UPDATED, updatedSubscriptionList);
            }
           
            try {
                if (getResourceListSize() == 0) {
                    logger.debug("[name=" + getName() + "] - waiting for new subscriptions ...");
                    bl.waitUntilFalse(0L);
                    logger.debug("[name=" + getName() + "] - waiting for new subscriptions ... done!");
                } else {
                    logger.debug("[name=" + getName() + "] - sleeping " + rate + " msec ...");
                    Thread.sleep(rate);
                    logger.debug("[name=" + getName() + "] - sleeping " + rate + " msec ... done!");
                }
            } catch (Throwable t) {
            }
        }
    }

    private void printNotification(final Notification notification, SensorEventArrayList eventList, String subscriptionId, String user) {
        if (notification == null) {
            return;
        }

        StringBuffer sb = new StringBuffer();
        sb.append("\n******* NOTIFICATION *******");
        sb.append("\nsubscriptionId: " + subscriptionId);
        sb.append("\nuser: " + user);
        sb.append("\nconsumer url: " + notification.getConsumerURL());
        if (notification.getTopic() != null && notification.getTopic().getName() != null) {
            sb.append("\topic name: " + notification.getTopic().getName());
        }

        /*
         * Event[] event = notification.getEvent();
         * 
         * if (event == null) { sb.append("\nNONE EVENT FOUND!"); }
         */

        if (eventList == null) {
            sb.append("\nNONE EVENT FOUND!");
        } else {
            SensorEvent event = null;

            for (int i = 0; i < eventList.size(); i++) {
                event = (SensorEvent) eventList.get(i);

                sb.append("\n").append(i).append(") event_name: ").append(event.getName());

                if (event.getTimestamp() != null) {
                    sb.append(" - timestamp: ").append(event.getTimestamp().getTime());
                }
                if (event.getExpirationTime() != null) {
                    sb.append(" - expirationTime: ").append(event.getExpirationTime().getTime());
                }

                sb.append(" - messages: ");

                String[] msg = event.getMessage();
                if (msg == null) {
                    sb.append("0");
                } else {
                    sb.append(msg.length);
                }

                /*
                 * if(msg != null) { for (int j = 0; j < msg.length; j++) {
                 * sb.append("\n\n******* EVENT " + i + "/" + event.length +
                 * "  MESSAGE " + j + "/" + msg.length + " *******");
                 * sb.append("\n" + msg[j]); sb.append("\n******* END MESSAGE "
                 * + j + " *******"); } }
                 */
            }
        }
        sb.append("\n******* END NOTIFICATION *******\n");
        logger.debug(sb.toString());
    }

    /*
     * private void printNotification(final Notification notification, String
     * subscriptionId, String user) { if (notification == null) { return; }
     * 
     * StringBuffer sb = new StringBuffer();
     * sb.append("\n******* NOTIFICATION *******");
     * sb.append("\nsubscriptionId: " + subscriptionId); sb.append("\nuser: " +
     * user); sb.append("\nconsumer url: " + notification.getConsumerURL()); if
     * (notification.getTopic() != null && notification.getTopic().getName() !=
     * null) { sb.append("\topic name: " + notification.getTopic().getName()); }
     * 
     * Event[] event = notification.getEvent();
     * 
     * if (event == null) { sb.append("\nNONE EVENT FOUND!"); return; }
     * 
     * for (int i = 0; i < event.length; i++) { if(event[i].getTimestamp() !=
     * null) { sb.append("\nevent timestamp: " +
     * event[i].getTimestamp().getTime()); } if(event[i].getProducer() != null)
     * { sb.append("\"event producer: " + event[i].getProducer()); }
     * 
     * String[] msg = event[i].getMessage();
     * 
     * if(msg != null) { for (int j = 0; j < msg.length; j++) {
     * sb.append("\n\n******* EVENT " + i + "/" + event.length + "  MESSAGE " +
     * j + "/" + msg.length + " *******"); sb.append("\n" + msg[j]);
     * sb.append("\n******* END MESSAGE " + j + " *******"); } } }
     * sb.append("\n\n******* END NOTIFICATION *******");
     * logger.debug(sb.toString()); }
     */


    private CEMonitorConsumerPort getCEMonitorConsumerPort(String consumerURL, SSLConfigSender sslConfigSender) throws IllegalArgumentException, MalformedURLException, ServiceException, NamingException {
        if (consumerURL == null) {
            throw new IllegalArgumentException("consumerURL not specified!");
        }

        if (sslConfigSender == null) {
            throw new IllegalArgumentException("sslConfigSender not specified!");
        }

        Handler transport = new SimpleTargetedChain(new SimpleChain(), sslConfigSender, new SimpleChain());
        SimpleProvider transportProvider = new SimpleProvider();
        transportProvider.deployTransport(HTTPTransport.DEFAULT_TRANSPORT_NAME, transport);

        CEMonitorConsumerLocator locator = new CEMonitorConsumerLocator(transportProvider);

        org.apache.axis.client.Stub stub = (org.apache.axis.client.Stub) locator.getCEMonitorConsumerPort(new URL(consumerURL));

        int milliseconds = 5000;

        if (serviceAttrs != null) {
            Attribute timeoutAttr = serviceAttrs.get("notificationTimeout_sec");
            if (timeoutAttr != null && timeoutAttr.size() > 0) {
                String timeout = (String) timeoutAttr.get(0);

                try {
                    milliseconds = Integer.parseInt(timeout);
                    if (milliseconds == 0) {
                        milliseconds = 1;
                        logger.error("[name=" + getName() + "] - illegal value error: the \"notificationTimeout_sec\" parameter value (" + timeout + ") must be >0. Using the default value (2 sec)");
                    }
                    milliseconds *= 1000;
                } catch (NumberFormatException e) {
                    logger.error("[name=" + getName() + "] - number format error: the \"notificationTimeout_sec\" parameter value (" + timeout + ") must have integer type type. Using the default value (1 sec)");
                }
            } else {
                logger.info("[name=" + getName() + "] - \"notificationTimeout_sec\" parameter not found!. Using the default value (2 sec)");
            }
        }

        stub.setTimeout(milliseconds);
        return (CEMonitorConsumerPort) stub;
    }

    private SSLConfigSender getSSLConfigSender(String credFile, String passphrase, String protocol) throws AxisFault, ServiceException, NamingException {
        if (serviceAttrs == null) {
            throw new NamingException("Missing service configuration");
        }

        Properties sslConfig = new java.util.Properties();
        // Attributes serviceAttrs = rootContext.getAttributes("");
        Attribute attr = null;

        if (credFile != null) {
            sslConfig.put(ContextWrapper.CREDENTIALS_PROXY_FILE, credFile);
        } else {
            attr = serviceAttrs.get("gridproxyfile");
            if (attr != null) {
                sslConfig.put(ContextWrapper.CREDENTIALS_PROXY_FILE, attr.get(0));
            } else {
                attr = serviceAttrs.get("sslcertfile");
                if (attr != null)
                    sslConfig.put(ContextWrapper.CREDENTIALS_CERT_FILE, attr.get(0));

                attr = serviceAttrs.get("sslkeyfile");
                if (attr != null)
                    sslConfig.put(ContextWrapper.CREDENTIALS_KEY_FILE, attr.get(0));

                attr = serviceAttrs.get("sslkeypasswd");
                if (attr != null)
                    sslConfig.put(ContextWrapper.CREDENTIALS_KEY_PASSWD, attr.get(0));
            }
        }

        if (passphrase != null) {
            sslConfig.put(ContextWrapper.CREDENTIALS_KEY_PASSWD, passphrase);
        } else {
            sslConfig.put(ContextWrapper.CREDENTIALS_KEY_PASSWD, "");
        }

        if (protocol != null)
            sslConfig.put(ContextWrapper.SSL_PROTOCOL, protocol);

        attr = serviceAttrs.get("sslCAfiles");

        if (attr != null) {
            sslConfig.put(ContextWrapper.CA_FILES, attr.get(0));
        }

        attr = serviceAttrs.get("sslCRLfiles");
        if (attr != null) {
            sslConfig.put(ContextWrapper.CRL_ENABLED, "true");
            sslConfig.put(ContextWrapper.CRL_FILES, attr.get(0));
            sslConfig.put(ContextWrapper.CRL_UPDATE_INTERVAL, "0s");

        } else {
            sslConfig.put(ContextWrapper.CRL_ENABLED, "false");
        }

        return new SSLConfigSender(sslConfig);
    }

    public void destroy() {
        exit = true;
        bl.setValue(false);

        super.destroy();
        
        logger.info("[name=" + getName() + "] - destroyed!");
    }
}
