/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Luigi Zangrando <zangrando@pd.infn.it>
 *
 */

package org.glite.ce.monitor.registry;

import java.io.File;
import java.io.FileNotFoundException;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NameClassPair;
import javax.naming.NamingException;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;

import org.glite.ce.commonj.CEResource;

import org.apache.log4j.Logger;

/**
 * This class is used to manage a list of homogeneous resources.<br> 
 * This list is initially specified in a configuration file 
 * and then can be modified adding or removing resources.
 */
public class ResourceRegistry {
    private DirContext rootCtx;

    private final static Logger logger = Logger.getLogger( ResourceRegistry.class.getName() );

    /**
     * Creates a new ResourceRegistry object and initializes it.
     * The configuration file name is set to the specified string.
     *
     * @param filename The name of the configuration file.
     *
     * @throws FileNotFoundException thrown if the configuration file named as specified has not been found.
     * @throws Exception DOCUMENT ME!
     */
    public ResourceRegistry(String cacheDir, int cacheSize, String factory) throws NamingException {
        File dir = new File(cacheDir);
        if(!dir.exists()) {
            dir.mkdirs();
        }
            
        Hashtable env = new Hashtable();
        env.put(Context.INITIAL_CONTEXT_FACTORY, factory);
        env.put(Context.PROVIDER_URL, cacheDir);
        env.put("cache.size", new Integer(cacheSize));

	    logger.debug("CTOR() - factory=["+factory+"] - cacheDir=["+cacheDir+"] - cacheSize=["+cacheSize+"]");

        rootCtx = (DirContext) (new InitialDirContext(env)).lookup("");        
    }
    

    /**
     * Add a new object in the resource registry. 
     * This object will be considered a persistent resource, and so it will be written in the repository. 
     * This new resource will be unequivocally identified in this registry by the specified key.
     *
     * @param key The resource unique identificator.
     * @param obj The new resource.
     *
     * @throws IllegalArgumentException, NamingException, NameAlreadyBoundException
     */
    public void register(String key, CEResource resource) throws IllegalArgumentException, NamingException, NameAlreadyBoundException {
        if (key == null) {
            throw (new IllegalArgumentException("The key argument is null"));
        }
        if (resource == null) {
            throw (new IllegalArgumentException("The resource argument is null"));
        }
                
        rootCtx.bind(key, resource);   
    }
    
    
    
    /**
     * Remove a resource from this registry specifying its identificator string.
     *
     * @param key The identificator string of the resource to be removed.
     * @throws NamingException 
     *
     * @throws Exception thrown if the specified key doesn't correspond to any registered resource.
     */
    public void unregister(String key) throws NamingException {
        if (key == null) {
            throw (new IllegalArgumentException("The key argument is null"));
        }
        
        rootCtx.unbind(key);   
    }
    

    public void update(String key, CEResource resource) throws IllegalArgumentException, NamingException, NameAlreadyBoundException {
        if (key == null) {
            throw (new IllegalArgumentException("The key argument is null"));
        }
        if (resource == null) {
            throw (new IllegalArgumentException("The resource argument is null"));
        }
                
        rootCtx.rebind(key, resource);   
    }
      
    
    
    
    /**
     * Get the resource specified by its identificator string.
     *
     * @param key The identificator string.
     *
     *
     * @return The searched resource if present in the registry, null otherwise.
     * @throws NamingException 
     */
    public Object getResource(String key) throws NamingException {
        Object obj = null;
        
        if (key != null) {
            obj = rootCtx.lookup(key);
        }

        return obj;
    }

    /**
     * Get all the resources present in this registry.
     *
     * @return The Collection of the registered resources or null if no resource has been registered yet.
     */
    public ArrayList getAllResources() {       
        ArrayList list = new ArrayList();
        
        try {   
            Enumeration allNames = rootCtx.list("");
            
            while (allNames.hasMoreElements()) {
                NameClassPair item = (NameClassPair) allNames.nextElement();
                
                list.add(rootCtx.lookup(item.getName()));
            }
        } catch (NamingException nEx) {
        }
        
        return list;
    }
}
