/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Authors: Luigi Zangrando <zangrando@pd.infn.it>
 *
 */

package org.glite.ce.monitor.registry;

import java.io.FileNotFoundException;
import java.net.URI;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import javax.naming.NamingException;

import org.apache.log4j.Logger;
import org.glite.ce.monitorapij.resource.types.Action;
import org.glite.ce.monitorapij.resource.types.Dialect;
import org.glite.ce.monitorapij.resource.types.Parameter;
import org.glite.ce.monitorapij.resource.types.Policy;
import org.glite.ce.monitorapij.resource.types.Property;
import org.glite.ce.monitorapij.resource.types.Query;
import org.glite.ce.monitorapij.resource.types.SubscriptionPersistent;
import org.glite.ce.monitorapij.resource.types.Topic;
import org.glite.ce.monitorapij.types.Subscription;


/**
 * 
 * This class is an extension of <code>ResourceRegistry</code> 
 * characterized to manage a list of subscriptions.<br> 
 * This list is initially specified in a configuration file 
 * and then can be modified adding or removing subscriptions.<br>
 * If the subscription is persistent, each modification on it will be saved in the configuration file.
 */
public class SubscriptionRegistry extends ResourceRegistry {
    private final static Logger logger = Logger.getLogger(SubscriptionRegistry.class.getName());
	
    /**
     * Creates a new SubscriptionRegistry object and initializes it.
     * The configuration file name is set to the specified string.
     *
     * @param filename The name of the configuration file.
     *
     * @throws FileNotFoundException thrown if the configuration file named as specified has not been found.
     * @throws Exception thrown if there is any problem writing or reading the configuration file.
     */
    public SubscriptionRegistry(String cacheDir, int cacheSize, String factory) throws FileNotFoundException, Exception {
        super(cacheDir, cacheSize, factory);        
        
        ArrayList resources = getAllResources();
        
        if(resources != null) {
            Iterator iterator = resources.iterator();
            
            while (iterator.hasNext()) {
                Object obj = iterator.next();
                
                if ((obj != null) && obj instanceof SubscriptionPersistent) {
                    SubscriptionPersistent subscription = (SubscriptionPersistent) obj;

                    if(subscription.removeOnStartup() || subscription.isExpired()) {
                        logger.debug("unregistering subscription: id = " + subscription.getId());
                        
                        unregister(subscription);
                    }                    
                }
            }            
        }
        
    }

    
    
    public SubscriptionPersistent makeSubscriptionPersistent(Subscription subscription) throws IllegalArgumentException, Exception {
        if (subscription == null) {
            throw (new IllegalArgumentException("the subscription is null"));
        }
        
        SubscriptionPersistent subscriptionPersistent = new SubscriptionPersistent();
        subscriptionPersistent.setExpirationTime(subscription.getExpirationTime());
        subscriptionPersistent.setType(subscription.getType());
        subscriptionPersistent.setMonitorConsumerURL(new URI(subscription.getMonitorConsumerURL().toString()));
        

        if(subscription.getJarPath() != null) {
            subscriptionPersistent.setJarPath(new URI(subscription.getJarPath().toString()));
        }
        
        org.glite.ce.monitorapij.types.Property[] property = subscription.getProperty();
        if(property != null) {
            Property[] newProperty = new Property[property.length];
            
            for(int i=0; i<property.length; i++) {
                newProperty[i] = new Property(property[i].getName(), property[i].getValue());
            }
            subscriptionPersistent.setProperty(newProperty);
        }        
        
        org.glite.ce.monitorapij.types.Topic topic = subscription.getTopic();
        org.glite.ce.monitorapij.types.Dialect[] dialect = topic.getDialect();
        
        Topic newtopic = new Topic(topic.getName());
        if(dialect != null) {
            Dialect[] newDialect = new Dialect[dialect.length];
            
            for(int i=0; i<dialect.length; i++) {
                newDialect[i] = new Dialect(dialect[i].getName(), dialect[i].getQueryLanguage());
            }
            newtopic.setDialect(newDialect);
        }
        
        subscriptionPersistent.setTopic(newtopic);
        org.glite.ce.monitorapij.types.Policy policy = subscription.getPolicy();
        
        if(policy != null) {            
            Policy newPolicy = new Policy();
            newPolicy.setRate(policy.getRate());
            
            org.glite.ce.monitorapij.types.Query query = policy.getQuery();            
            if(query != null) {
                newPolicy.setQuery(new Query(query.getExpression(), query.getQueryLanguage()));
            }
            
            org.glite.ce.monitorapij.types.Action[] action = policy.getAction();            
            if(action != null) {
                Action[] newAction = new Action[action.length];
                
                for(int i=0; i<action.length; i++) {
                    newAction[i] = new Action(action[i].getName());
                    newAction[i].setDoActionWhenQueryIs(action[i].isDoActionWhenQueryIs());        
                    
                    org.glite.ce.monitorapij.types.Parameter[] parameter = action[i].getParameter();
                    if(parameter != null) {
                        Parameter[] newParameter = new Parameter[parameter.length];

                        for(int j=0; j<parameter.length; j++) {
                            newParameter[j] = new Parameter(parameter[j].getName(), parameter[j].getValue());
                        }
                        newAction[i].setParameter(newParameter);
                    }
                }
                
                newPolicy.setAction(newAction);
            }
            subscriptionPersistent.setPolicy(newPolicy);
        }

        return subscriptionPersistent;
    }
    

    
    public Subscription makeSubscription(SubscriptionPersistent subscriptionPersistent) throws IllegalArgumentException, Exception {
        if (subscriptionPersistent == null) {
            throw (new IllegalArgumentException("the subscription is null"));
        }
        
        Subscription subscription = new Subscription();
        subscription.setId(subscriptionPersistent.getId());
        subscription.setName(subscriptionPersistent.getName());
        subscription.setExpirationTime(subscriptionPersistent.getExpirationTime());
        subscription.setType(subscriptionPersistent.getType());
        subscription.setMonitorConsumerURL(new org.apache.axis.types.URI(subscriptionPersistent.getMonitorConsumerURL().toString()));
        
        if(subscriptionPersistent.getJarPath() != null) {
            subscription.setJarPath(new org.apache.axis.types.URI(subscriptionPersistent.getJarPath().toString()));
        }
        
        Property[] property = subscriptionPersistent.getProperty();
        if(property != null) {
            org.glite.ce.monitorapij.types.Property[] newProperty = new org.glite.ce.monitorapij.types.Property[property.length];
            
            for(int i=0; i<property.length; i++) {
                newProperty[i] = new org.glite.ce.monitorapij.types.Property(property[i].getName(), property[i].getValue());
            }
            subscription.setProperty(newProperty);
        }   
        
        Topic topic = subscriptionPersistent.getTopic();
        
        org.glite.ce.monitorapij.types.Topic newtopic = new org.glite.ce.monitorapij.types.Topic();
        newtopic.setName(topic.getName());

        Dialect[] dialect = topic.getDialect();
        if(dialect != null) {
            org.glite.ce.monitorapij.types.Dialect[] newDialect = new org.glite.ce.monitorapij.types.Dialect[dialect.length];
            
            for(int i=0; i<dialect.length; i++) {
                newDialect[i] = new org.glite.ce.monitorapij.types.Dialect();
                newDialect[i].setName(dialect[i].getName());
                newDialect[i].setQueryLanguage(dialect[i].getQueryLanguage());
            }
            newtopic.setDialect(newDialect);
        }
        
        subscription.setTopic(newtopic);
        Policy policy = subscriptionPersistent.getPolicy();
        
        if(policy != null) {            
            org.glite.ce.monitorapij.types.Policy newPolicy = new org.glite.ce.monitorapij.types.Policy();
            newPolicy.setRate(policy.getRate());

            Query query = policy.getQuery();
            if(query != null) {
                newPolicy.setQuery(new org.glite.ce.monitorapij.types.Query(query.getExpression(), query.getQueryLanguage()));
            }
	    
            Action[] action = policy.getAction();            
            if(action != null) {
                org.glite.ce.monitorapij.types.Action[] newAction = new org.glite.ce.monitorapij.types.Action[action.length];
                
                for(int i=0; i<action.length; i++) {
                    newAction[i] = new org.glite.ce.monitorapij.types.Action();
                    newAction[i].setName(action[i].getName());
                    newAction[i].setDoActionWhenQueryIs(action[i].isDoActionWhenQueryIs());    
                    
                    Parameter[] parameter = action[i].getParameter();
                    if(parameter != null) {
                        org.glite.ce.monitorapij.types.Parameter[] newParameter = new org.glite.ce.monitorapij.types.Parameter[parameter.length];

                        for(int j=0; j<parameter.length; j++) {
                            newParameter[j] = new org.glite.ce.monitorapij.types.Parameter(parameter[j].getName(), (String)parameter[j].getValue());
                        }
                        newAction[i].setParameter(newParameter);
                    }
                }
                
                newPolicy.setAction(newAction);
            }
            subscription.setPolicy(newPolicy);
        }

        return subscription;
    }
    
    
    
    /**
     * Get a <code>SubscriptionPersistent</code> object created using the specified subscription.<br>
     * This will be added in the registry. as this is considered persistent by default. 
     *
     * @param subscription The subscription to add.
     *
     * @return A <code>SubscriptionPersistent</code> generated using the specified subscription.
     *
     * @throws Exception thrown if the subscription is null or expired or 
     * if some problem occurs reading or writing the configuration file.
     */
    public void register(SubscriptionPersistent subscription) throws IllegalArgumentException, Exception {
        if (subscription == null) {
            throw (new IllegalArgumentException("the subscription is null"));
        }
        
        register(subscription.getId(), subscription);        
    }

    public void unregister(SubscriptionPersistent subscription) throws IllegalArgumentException, Exception {
        if (subscription == null) {
            throw (new IllegalArgumentException("the subscription is null"));
        }
        
        unregister(subscription.getId());        
    }

    

    public void update(SubscriptionPersistent subscription) throws Exception {
        if (subscription == null) {
            throw (new Exception("the subscription is null"));
        }
 
        update(subscription.getId(), subscription);
    }
    
    
    /**
     * Get the registered <code>SubscriptionPersistent</code> specifying its ID.
     *
     * @param id The identificator of the subscription.
     *
     * @return The required <code>SubscriptionPersistent</code> if present in the registry, 
     * null otherwise.
     * @throws NamingException 
     */
    public SubscriptionPersistent getSubscription(String id) throws NamingException {
        Object subscription = getResource(id);

        if ((subscription != null) && subscription instanceof SubscriptionPersistent) {
            return (SubscriptionPersistent) subscription;
        } else {
            return null;
        }
    }
        

    public ArrayList getSubscriptionBySubscriberId(String subscriberId) {
        if (subscriberId == null) {
            return null;
        }

        ArrayList resources = getAllResources();
        if (resources == null) {
            return null;
        }

        ArrayList goodSubscriptions = new ArrayList();

        Iterator iterator = resources.iterator();

        while (iterator.hasNext()) {
            Object obj = iterator.next();

            if ((obj != null) && obj instanceof SubscriptionPersistent) {
                SubscriptionPersistent subscription = (SubscriptionPersistent) obj;

                if (subscription.isExpired()) {
                    try {
                        unregister(subscription.getId());
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                } else {
                    String sub_Id = subscription.getSubscriberId();

                    if ((sub_Id != null) && sub_Id.equals(subscriberId)) {
                        goodSubscriptions.add(subscription);
                    }
                }
            }
        }

       
        return goodSubscriptions;
    }

    
    

    public SubscriptionPersistent[] getAllSubscriptions() {
        Collection resources = getAllResources();

        if (resources == null) {
            return null;
        }
        
        SubscriptionPersistent[] result = new SubscriptionPersistent[resources.size()];
        result = (SubscriptionPersistent[])resources.toArray(result);
        
        return result;
    }
    
    
    
    /**
     * Get the registered <code>SubscriptionPersistent</code> having the specified <code>Topic</code>.<br>
     * Expired subscriptions are filtered and removed from the registry. 
     *
     * @param topic The <code>Topic</code> characterizing the required subscriptions.
     *
     * @return An array of <code>SubscriptionPersistent</code> containing the required subscriptions, if existing.
     */
    public ArrayList getSubscriptionByTopic(Topic topic) {
        if (topic == null) {
            return null;
        }

        Collection resources = getAllResources();

        if (resources == null) {
            return null;
        }

        ArrayList goodSubscriptions = new ArrayList();

        Iterator iterator = resources.iterator();

        while (iterator.hasNext()) {
            Object obj = iterator.next();

            if ((obj != null) && obj instanceof SubscriptionPersistent) {
                SubscriptionPersistent subscription = (SubscriptionPersistent) obj;

                if (subscription.isExpired()) {
                    try {
                        unregister(subscription.getId());
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                } else {
                    if ((subscription.getTopic() != null) && subscription.getTopic().getName().equals(topic.getName())) {
                        goodSubscriptions.add(subscription);
                    }
                }
            }
        }

        return goodSubscriptions;
    }
}
